import uuid
from datetime import datetime

from sqlmodel import Field, SQLModel


class TaggedItem(SQLModel, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    tag_id: uuid.UUID = Field(foreign_key="tag.id")
    item_id: uuid.UUID
    item_type: str


class Tag(SQLModel, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    name: str = Field(index=True)
    site_id: uuid.UUID = Field(index=True, foreign_key="site.id", ondelete="CASCADE")

class PostBase(SQLModel):
    title: str | None = Field(default=None, max_length=255)
    slug: str = Field(index=True, unique=True)

    created_at: datetime = Field(default_factory=datetime.now, nullable=False)
    updated_at: datetime = Field(default_factory=datetime.now, nullable=False)
    author: str | None = Field(default=None)
    site_id: uuid.UUID = Field(index=True, foreign_key="site.id", ondelete="CASCADE")
    # content: str | None = Field(default=None)


class Post(PostBase, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)


class BlogPostItem(PostBase):
    id: uuid.UUID


class BlogPostListResponse(SQLModel):
    data: list[BlogPostItem]
    count: int


class BlogPostCreateReq(SQLModel):
    content: str | None = None
    title: str | None = None
    tags: list[str] = []
    slug: str | None = None
    siteId: uuid.UUID


class BlogPostCreateResponse(SQLModel):
    id: uuid.UUID
    # siteId: uuid.UUID


class BlogPostUpdateRequest(SQLModel):
    id: uuid.UUID
    title: str
    content: str
    slug: str | None = None
    tags: list[str] = []



class BlogPostUpdateResponse(SQLModel):
    id: uuid.UUID



class BlogPostDetailResponse(SQLModel):
    id: uuid.UUID
    title: str
    content: str
    tags: list[str]
    created_at: datetime
    updated_at: datetime
    author: str | None


class PostContent(SQLModel, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    post_id: uuid.UUID = Field(foreign_key="post.id", unique=True, ondelete="CASCADE")
    content: str = Field(default=None)  # Full content


# 练习 tags 与 vedio 的关系
class VideoBase(SQLModel):
    title: str = Field(index=True)
    description: str | None = Field(default=None)
    url: str
    duration: int  # in seconds
    created_at: datetime = Field(default_factory=datetime.now, nullable=False)
    updated_at: datetime = Field(default_factory=datetime.now, nullable=False)
    author: str | None = Field(default=None)


class Video(VideoBase, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)


class TagItemPublic(SQLModel):
    name: str


class TagListResponse(SQLModel):
    data: list[TagItemPublic]
    count: int



from datetime import datetime
from typing import List, Optional

from autogen_agentchat.agents import AssistantAgent, UserProxyAgent
from autogen_agentchat.conditions import MaxMessageTermination, TextMentionTermination
from autogen_agentchat.teams import RoundRobinGroupChat, SelectorGroupChat
from autogen_core import ComponentModel
from autogen_core.tools import FunctionTool
from autogen_ext.agents.web_surfer import MultimodalWebSurfer

from ..agents.ag.model_client import get_oai_Model
from ..models.ag import Gallery, GalleryComponents, GalleryItems, GalleryMetadata


class GalleryBuilder:
    """Builder class for creating AutoGen component galleries."""

    def __init__(self, id: str, name: str):
        self.id = id
        self.name = name
        self.url: Optional[str] = None
        self.teams: List[ComponentModel] = []
        self.agents: List[ComponentModel] = []
        self.models: List[ComponentModel] = []
        self.tools: List[ComponentModel] = []
        self.terminations: List[ComponentModel] = []

        # Default metadata
        self.metadata = GalleryMetadata(
            author="AutoGen Team",
            created_at=datetime.now(),
            updated_at=datetime.now(),
            version="1.0.0",
            description="",
            tags=[],
            license="MIT",
            category="conversation",
        )

    def set_metadata(
        self,
        author: Optional[str] = None,
        version: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        license: Optional[str] = None,
        category: Optional[str] = None,
    ) -> "GalleryBuilder":
        """Update gallery metadata."""
        if author:
            self.metadata.author = author
        if version:
            self.metadata.version = version
        if description:
            self.metadata.description = description
        if tags:
            self.metadata.tags = tags
        if license:
            self.metadata.license = license
        if category:
            self.metadata.category = category
        return self

    def add_team(self, team: ComponentModel) -> "GalleryBuilder":
        """Add a team component to the gallery."""
        self.teams.append(team)
        return self

    def add_agent(self, agent: ComponentModel) -> "GalleryBuilder":
        """Add an agent component to the gallery."""
        self.agents.append(agent)
        return self

    def add_model(self, model: ComponentModel) -> "GalleryBuilder":
        """Add a model component to the gallery."""
        self.models.append(model)
        return self

    def add_tool(self, tool: ComponentModel) -> "GalleryBuilder":
        """Add a tool component to the gallery."""
        self.tools.append(tool)
        return self

    def add_termination(self, termination: ComponentModel) -> "GalleryBuilder":
        """Add a termination condition component to the gallery."""
        self.terminations.append(termination)
        return self

    def build(self) -> Gallery:
        """Build and return the complete gallery."""
        # Update timestamps
        self.metadata.updated_at = datetime.now()

        return Gallery(
            id=self.id,
            name=self.name,
            url=self.url,
            metadata=self.metadata,
            items=GalleryItems(
                teams=self.teams,
                components=GalleryComponents(
                    agents=self.agents,
                    models=self.models,
                    tools=self.tools,
                    terminations=self.terminations,
                ),
            ),
        )


def create_default_gallery() -> Gallery:
    """Create a default gallery with all components including calculator and web surfer teams."""
    builder = GalleryBuilder(id="gallery_default", name="Default Component Gallery")

    # Set metadata
    builder.set_metadata(
        description="A default gallery containing basic components for human-in-loop conversations",
        tags=["human-in-loop", "assistant"],
        category="conversation",
    )

    # Create base model client
    # base_model = OpenAIChatCompletionClient(model="gpt-4o-mini")
    base_model = get_oai_Model()
    # base_model = OpenAIChatCompletionClient(
    #     model="deepseek-r1:1.5b",
    #     base_url="http://localhost:11434/v1",
    #     api_key="placeholder",
    #     model_info={
    #         "vision": False,
    #         "function_calling": ,
    #         "json_output": False,
    #         "family": ModelFamily.R1,
    #     },
    # )
    builder.add_model(base_model.dump_component())

    # mistral_vllm_model = OpenAIChatCompletionClient(
    #     model="TheBloke/Mistral-7B-Instruct-v0.2-GGUF",
    #     base_url="http://localhost:1234/v1",
    #     model_info=ModelInfo(vision=False, function_calling=True, json_output=False),
    # )
    # builder.add_model(mistral_vllm_model.dump_component())

    # Create websurfer model client
    # websurfer_model = OpenAIChatCompletionClient(model="gpt-4o-mini")
    # builder.add_model(websurfer_model.dump_component())

    def calculator(a: float, b: float, operator: str) -> str:
        try:
            if operator == "+":
                return str(a + b)
            elif operator == "-":
                return str(a - b)
            elif operator == "*":
                return str(a * b)
            elif operator == "/":
                if b == 0:
                    return "Error: Division by zero"
                return str(a / b)
            else:
                return "Error: Invalid operator. Please use +, -, *, or /"
        except Exception as e:
            return f"Error: {str(e)}"

    # Create calculator tool
    calculator_tool = FunctionTool(
        name="calculator",
        description="A simple calculator that performs basic arithmetic operations",
        func=calculator,
        global_imports=[],
    )
    # calculator_tool.
    builder.add_tool(calculator_tool.dump_component())

    # Create termination conditions for calculator team
    calc_text_term = TextMentionTermination(text="TERMINATE")
    calc_max_term = MaxMessageTermination(max_messages=10)
    calc_or_term = calc_text_term | calc_max_term

    builder.add_termination(calc_text_term.dump_component())
    builder.add_termination(calc_max_term.dump_component())
    builder.add_termination(calc_or_term.dump_component())

    # Create calculator assistant agent
    calc_assistant = AssistantAgent(
        name="assistant_agent",
        system_message="You are a helpful assistant. Solve tasks carefully. When done, say TERMINATE.",
        model_client=base_model,
        tools=[calculator_tool],
    )
    builder.add_agent(calc_assistant.dump_component())

    # Create calculator team
    calc_team = RoundRobinGroupChat(
        participants=[calc_assistant], termination_condition=calc_or_term
    )
    builder.add_team(calc_team.dump_component())

    # Create web surfer agent
    websurfer_agent = MultimodalWebSurfer(
        name="websurfer_agent",
        description="an agent that solves tasks by browsing the web",
        model_client=base_model,
        headless=True,
    )
    builder.add_agent(websurfer_agent.dump_component())

    # Create web surfer verification assistant
    verification_assistant = AssistantAgent(
        name="assistant_agent",
        description="an agent that verifies and summarizes information",
        system_message="You are a task verification assistant who is working with a web surfer agent to solve tasks. At each point, check if the task has been completed as requested by the user. If the websurfer_agent responds and the task has not yet been completed, respond with what is left to do and then say 'keep going'. If and only when the task has been completed, summarize and present a final answer that directly addresses the user task in detail and then respond with TERMINATE.",
        model_client=base_model,
    )
    builder.add_agent(verification_assistant.dump_component())

    # Create web surfer user proxy
    web_user_proxy = UserProxyAgent(
        name="user_proxy",
        description="a human user that should be consulted only when the assistant_agent is unable to verify the information provided by the websurfer_agent",
    )
    builder.add_agent(web_user_proxy.dump_component())

    # Create web surfer team termination conditions
    web_max_term = MaxMessageTermination(max_messages=20)
    web_text_term = TextMentionTermination(text="TERMINATE")
    web_termination = web_max_term | web_text_term
    builder.add_termination(web_termination.dump_component())

    # Create web surfer team
    selector_prompt = """You are the cordinator of role play game. The following roles are available:
{roles}. Given a task, the websurfer_agent will be tasked to address it by browsing the web and providing information.  The assistant_agent will be tasked with verifying the information provided by the websurfer_agent and summarizing the information to present a final answer to the user. If the task  needs assistance from a human user (e.g., providing feedback, preferences, or the task is stalled), you should select the user_proxy role to provide the necessary information.
Read the following conversation. Then select the next role from {participants} to play. Only return the role.

{history}

Read the above conversation. Then select the next role from {participants} to play. Only return the role."""

    websurfer_team = SelectorGroupChat(
        participants=[websurfer_agent, verification_assistant, web_user_proxy],
        selector_prompt=selector_prompt,
        model_client=base_model,
        termination_condition=web_termination,
    )
    builder.add_team(websurfer_team.dump_component())

    return builder.build()


# if __name__ == "__main__":
#     # Create and save the gallery
#     gallery = create_default_gallery()

#     # Print as JSON
#     print(gallery.model_dump_json(indent=2))

#     # Save to file
#     with open("gallery_default.json", "w") as f:
#         f.write(gallery.model_dump_json(indent=2))
#         f.write(gallery.model_dump_json(indent=2))



from bs4 import BeautifulSoup


def extract_title_from_html(html_content: str) -> str:
    """
    从html 中提取title
    """
    # 使用BeautifulSoup解析HTML内容
    soup = BeautifulSoup(html_content, "html.parser")

    # 查找title标签
    title_tag = soup.find("title")
    if title_tag:
        return title_tag.text.strip()

    # 如果没有title标签，查找h1标签
    h1_tag = soup.find("h1")
    if h1_tag:
        return h1_tag.text.strip()

    # 如果没有h1标签，查找h2标签
    h2_tag = soup.find("h2")
    if h2_tag:
        return h2_tag.text.strip()

    # 如果都没有找到，返回空字符串
    return ""



import os
from pathlib import Path

import gradio as gr
from dotenv import load_dotenv
from fastapi import FastAPI
from huggingface_hub import InferenceClient

default_env_files = [".env", ".env.local", "../gomtm/env/dev.env"]

load_dotenv()
# 搜索并加载环境变量文件
for env_file in default_env_files:
    env_path = Path(env_file)
    if env_path.exists():
        load_dotenv(env_path)
# token = os.getenv("HUGGINGFACEHUB_API_TOKEN")
# if not token:
#     raise ValueError("HUGGINGFACEHUB_API_TOKEN 未设置")


def getInferenceClient():
    """
    For more information on `huggingface_hub` Inference API support, please check the docs: https://huggingface.co/docs/huggingface_hub/v0.22.2/en/guides/inference
    """
    if not os.getenv("HUGGINGFACEHUB_API_TOKEN"):
        raise ValueError("HUGGINGFACEHUB_API_TOKEN 未设置")
    return InferenceClient(
        "HuggingFaceH4/zephyr-7b-beta",
        token=os.getenv("HUGGINGFACEHUB_API_TOKEN"),
    )


def respond(
    message,
    history: list[tuple[str, str]],
    system_message,
    max_tokens,
    temperature,
    top_p,
):
    messages = [{"role": "system", "content": system_message}]

    for val in history:
        if val[0]:
            messages.append({"role": "user", "content": val[0]})
        if val[1]:
            messages.append({"role": "assistant", "content": val[1]})

    messages.append({"role": "user", "content": message})

    response = ""

    for message in getInferenceClient().chat_completion(
        messages,
        max_tokens=max_tokens,
        stream=True,
        temperature=temperature,
        top_p=top_p,
    ):
        token = message.choices[0].delta.content

        response += token
        yield response


def greet(name):
    return "Hello greeter " + name + "!"


# 新增一个示例函数
def calculate(a: int, b: int, operation: str):
    if operation == "add":
        return a + b
    elif operation == "multiply":
        return a * b
    return None


with gr.Blocks() as demo:
    # 聊天界面
    chatbot = gr.ChatInterface(
        respond,
        additional_inputs=[
            gr.Textbox(value="You are a friendly Chatbot.", label="System message222"),
            gr.Slider(
                minimum=1, maximum=2048, value=512, step=1, label="Max new tokens"
            ),
            gr.Slider(
                minimum=0.1, maximum=4.0, value=0.7, step=0.1, label="Temperature"
            ),
            gr.Slider(
                minimum=0.1,
                maximum=1.0,
                value=0.95,
                step=0.05,
                label="Top-p (nucleus sampling)",
            ),
        ],
    )

    # 添加 greet API
    gr.Interface(
        fn=greet,
        inputs=gr.Textbox(label="Name"),
        outputs=gr.Textbox(label="Output"),
        api_name="greet",  # 这将创建 /greet 端点
    )

    # 添加 calculate API
    gr.Interface(
        fn=calculate,
        inputs=[
            gr.Number(label="First Number"),
            gr.Number(label="Second Number"),
            gr.Radio(choices=["add", "multiply"], label="Operation"),
        ],
        outputs=gr.Number(label="Result"),
        api_name="calculate",  # 这将创建 /calculate 端点
    )


def mount_gradio_app(app: FastAPI):
    """
    将 gradio 应用挂载到 fastapi 的根
    """
    import gradio as gr

    from .gradio_app import demo

    gr.mount_gradio_app(app, demo, path="/")


if __name__ == "__main__":
    demo.launch(
        share=True,
        server_name="0.0.0.0",
        server_port=18089,
    )



from fastapi import status


class SkyvernException(Exception):
    def __init__(self, message: str | None = None):
        self.message = message
        super().__init__(message)


class SkyvernHTTPException(SkyvernException):
    def __init__(self, message: str | None = None, status_code: int = status.HTTP_400_BAD_REQUEST):
        self.status_code = status_code
        super().__init__(message)


class DisabledBlockExecutionError(SkyvernHTTPException):
    def __init__(self, message: str | None = None):
        super().__init__(message, status_code=status.HTTP_400_BAD_REQUEST)


class InvalidOpenAIResponseFormat(SkyvernException):
    def __init__(self, message: str | None = None):
        super().__init__(f"Invalid response format: {message}")


class FailedToSendWebhook(SkyvernException):
    def __init__(
        self,
        task_id: str | None = None,
        workflow_run_id: str | None = None,
        workflow_id: str | None = None,
    ):
        workflow_run_str = f"workflow_run_id={workflow_run_id}" if workflow_run_id else ""
        workflow_str = f"workflow_id={workflow_id}" if workflow_id else ""
        task_str = f"task_id={task_id}" if task_id else ""
        super().__init__(f"Failed to send webhook. {workflow_run_str} {workflow_str} {task_str}")


class ProxyLocationNotSupportedError(SkyvernException):
    def __init__(self, proxy_location: str | None = None):
        super().__init__(f"Unknown proxy location: {proxy_location}")


class TaskNotFound(SkyvernHTTPException):
    def __init__(self, task_id: str | None = None):
        super().__init__(f"Task {task_id} not found", status_code=status.HTTP_404_NOT_FOUND)


class ScriptNotFound(SkyvernException):
    def __init__(self, script_name: str | None = None):
        super().__init__(f"Script {script_name} not found. Has the script been registered?")


class MissingElement(SkyvernException):
    def __init__(self, selector: str | None = None, element_id: str | None = None):
        super().__init__(
            f"Found no elements. Might be due to previous actions which removed this element."
            f" selector={selector} element_id={element_id}",
        )


class MultipleElementsFound(SkyvernException):
    def __init__(self, num: int, selector: str | None = None, element_id: str | None = None):
        super().__init__(
            f"Found {num} elements. Expected 1. num_elements={num} selector={selector} element_id={element_id}",
        )


class MissingFileUrl(SkyvernException):
    def __init__(self) -> None:
        super().__init__("File url is missing.")


class ImaginaryFileUrl(SkyvernException):
    def __init__(self, file_url: str) -> None:
        super().__init__(f"File url {file_url} is imaginary.")


class MissingBrowserState(SkyvernException):
    def __init__(self, task_id: str) -> None:
        super().__init__(f"Browser state for task {task_id} is missing.")


class MissingBrowserStatePage(SkyvernException):
    def __init__(self, task_id: str | None = None, workflow_run_id: str | None = None):
        task_str = f"task_id={task_id}" if task_id else ""
        workflow_run_str = f"workflow_run_id={workflow_run_id}" if workflow_run_id else ""
        super().__init__(f"Browser state page is missing. {task_str} {workflow_run_str}")


class MissingWorkflowRunBrowserState(SkyvernException):
    def __init__(self, workflow_run_id: str, task_id: str) -> None:
        super().__init__(f"Browser state for workflow run {workflow_run_id} and task {task_id} is missing.")


class CaptchaNotSolvedInTime(SkyvernException):
    def __init__(self, task_id: str, final_state: str) -> None:
        super().__init__(f"Captcha not solved in time for task {task_id}. Final state: {final_state}")


class EnablingCaptchaSolver(SkyvernException):
    def __init__(self) -> None:
        super().__init__("Enabling captcha solver. Reload the page and try again.")


class ContextParameterValueNotFound(SkyvernException):
    def __init__(self, parameter_key: str, existing_keys: list[str], workflow_run_id: str) -> None:
        super().__init__(
            f"Context parameter value not found during workflow run {workflow_run_id}. "
            f"Parameter key: {parameter_key}. Existing keys: {existing_keys}"
        )


class UnknownBlockType(SkyvernException):
    def __init__(self, block_type: str) -> None:
        super().__init__(f"Unknown block type {block_type}")


class WorkflowNotFound(SkyvernHTTPException):
    def __init__(
        self,
        workflow_id: str | None = None,
        workflow_permanent_id: str | None = None,
        version: int | None = None,
    ) -> None:
        workflow_repr = ""
        if workflow_id:
            workflow_repr = f"workflow_id={workflow_id}"
        if workflow_permanent_id:
            if version:
                workflow_repr = f"workflow_permanent_id={workflow_permanent_id}, version={version}"
            else:
                workflow_repr = f"workflow_permanent_id={workflow_permanent_id}"

        super().__init__(
            f"Workflow not found. {workflow_repr}",
            status_code=status.HTTP_404_NOT_FOUND,
        )


class WorkflowRunNotFound(SkyvernException):
    def __init__(self, workflow_run_id: str) -> None:
        super().__init__(f"WorkflowRun {workflow_run_id} not found")


class MissingValueForParameter(SkyvernHTTPException):
    def __init__(self, parameter_key: str, workflow_id: str, workflow_run_id: str) -> None:
        super().__init__(
            f"Missing value for parameter {parameter_key} in workflow run {workflow_run_id} of workflow {workflow_id}"
        )


class WorkflowParameterNotFound(SkyvernHTTPException):
    def __init__(self, workflow_parameter_id: str) -> None:
        super().__init__(
            f"Workflow parameter {workflow_parameter_id} not found",
            status_code=status.HTTP_404_NOT_FOUND,
        )


class FailedToNavigateToUrl(SkyvernException):
    def __init__(self, url: str, error_message: str) -> None:
        self.url = url
        self.error_message = error_message
        super().__init__(f"Failed to navigate to url {url}. Error message: {error_message}")


class FailedToReloadPage(SkyvernException):
    def __init__(self, url: str, error_message: str) -> None:
        self.url = url
        self.error_message = error_message
        super().__init__(f"Failed to reload page url {url}. Error message: {error_message}")


class FailedToStopLoadingPage(SkyvernException):
    def __init__(self, url: str, error_message: str) -> None:
        self.url = url
        self.error_message = error_message
        super().__init__(f"Failed to stop loading page url {url}. Error message: {error_message}")


class UnexpectedTaskStatus(SkyvernException):
    def __init__(self, task_id: str, status: str) -> None:
        super().__init__(f"Unexpected task status {status} for task {task_id}")


class InvalidWorkflowTaskURLState(SkyvernException):
    def __init__(self, workflow_run_id: str) -> None:
        super().__init__(f"No Valid URL found in the first task of workflow run {workflow_run_id}")


class DisabledFeature(SkyvernException):
    def __init__(self, feature: str) -> None:
        super().__init__(f"Feature {feature} is disabled")


class UnknownBrowserType(SkyvernException):
    def __init__(self, browser_type: str) -> None:
        super().__init__(f"Unknown browser type {browser_type}")


class UnknownErrorWhileCreatingBrowserContext(SkyvernException):
    def __init__(self, browser_type: str, exception: Exception) -> None:
        super().__init__(
            f"Unknown error while creating browser context for {browser_type}. Exception type: {type(exception)} Exception message: {str(exception)}"
        )


class BrowserStateMissingPage(SkyvernException):
    def __init__(self) -> None:
        super().__init__("BrowserState is missing the main page")


class OrganizationNotFound(SkyvernHTTPException):
    def __init__(self, organization_id: str) -> None:
        super().__init__(
            f"Organization {organization_id} not found",
            status_code=status.HTTP_404_NOT_FOUND,
        )


class StepNotFound(SkyvernHTTPException):
    def __init__(self, organization_id: str, task_id: str, step_id: str | None = None) -> None:
        super().__init__(
            f"Step {step_id or 'latest'} not found. organization_id={organization_id} task_id={task_id}",
            status_code=status.HTTP_404_NOT_FOUND,
        )


class FailedToTakeScreenshot(SkyvernException):
    def __init__(self, error_message: str) -> None:
        super().__init__(f"Failed to take screenshot. Error message: {error_message}")


class EmptyScrapePage(SkyvernException):
    def __init__(self) -> None:
        super().__init__("Failed to scrape the page, returned an NONE result")


class WorkflowRunContextNotInitialized(SkyvernException):
    def __init__(self, workflow_run_id: str) -> None:
        super().__init__(f"WorkflowRunContext not initialized for workflow run {workflow_run_id}")


class DownloadFileMaxSizeExceeded(SkyvernException):
    def __init__(self, max_size: int) -> None:
        self.max_size = max_size
        super().__init__(f"Download file size exceeded the maximum allowed size of {max_size} MB.")


class BitwardenBaseError(SkyvernException):
    def __init__(self, message: str) -> None:
        super().__init__(f"Bitwarden error: {message}")


class BitwardenLoginError(BitwardenBaseError):
    def __init__(self, message: str) -> None:
        super().__init__(f"Error logging in to Bitwarden: {message}")


class BitwardenUnlockError(BitwardenBaseError):
    def __init__(self, message: str) -> None:
        super().__init__(f"Error unlocking Bitwarden: {message}")


class BitwardenListItemsError(BitwardenBaseError):
    def __init__(self, message: str) -> None:
        super().__init__(f"Error listing items in Bitwarden: {message}")


class BitwardenTOTPError(BitwardenBaseError):
    def __init__(self, message: str) -> None:
        super().__init__(f"Error generating TOTP in Bitwarden: {message}")


class BitwardenLogoutError(BitwardenBaseError):
    def __init__(self, message: str) -> None:
        super().__init__(f"Error logging out of Bitwarden: {message}")


class BitwardenSyncError(BitwardenBaseError):
    def __init__(self, message: str) -> None:
        super().__init__(f"Error syncing Bitwarden: {message}")


class BitwardenAccessDeniedError(BitwardenBaseError):
    def __init__(self) -> None:
        super().__init__(
            "Current organization does not have access to the specified Bitwarden collection. "
            "Contact Skyvern support to enable access. This is a security layer on top of Bitwarden, "
            "Skyvern team needs to let your Skyvern account access the Bitwarden collection."
        )


class UnknownElementTreeFormat(SkyvernException):
    def __init__(self, fmt: str) -> None:
        super().__init__(f"Unknown element tree format {fmt}")


class StepTerminationError(SkyvernException):
    def __init__(self, step_id: str, reason: str) -> None:
        super().__init__(f"Step {step_id} cannot be executed and task is failed. Reason: {reason}")


class StepUnableToExecuteError(SkyvernException):
    def __init__(self, step_id: str, reason: str) -> None:
        super().__init__(f"Step {step_id} cannot be executed and task execution is stopped. Reason: {reason}")


class SVGConversionFailed(SkyvernException):
    def __init__(self, svg_html: str) -> None:
        super().__init__(f"Failed to convert SVG after max retries. svg_html={svg_html}")


class UnsupportedActionType(SkyvernException):
    def __init__(self, action_type: str):
        super().__init__(f"Unsupport action type: {action_type}")


class InvalidElementForTextInput(SkyvernException):
    def __init__(self, element_id: str, tag_name: str):
        super().__init__(f"The {tag_name} element with id={element_id} doesn't support text input.")


class ElementIsNotLabel(SkyvernException):
    def __init__(self, tag_name: str):
        super().__init__(f"<{tag_name}> element is not <label>")


class NoneFrameError(SkyvernException):
    def __init__(self, frame_id: str):
        super().__init__(f"frame content is none. frame_id={frame_id}")


class MissingElementDict(SkyvernException):
    def __init__(self, element_id: str) -> None:
        super().__init__(f"Invalid element id. element_id={element_id}")


class MissingElementInIframe(SkyvernException):
    def __init__(self, element_id: str) -> None:
        super().__init__(f"Found no iframe includes the element. element_id={element_id}")


class MissingElementInCSSMap(SkyvernException):
    def __init__(self, element_id: str) -> None:
        super().__init__(f"Found no css selector in the CSS map for the element. element_id={element_id}")


class InputActionOnSelect2Dropdown(SkyvernException):
    def __init__(self, element_id: str):
        super().__init__(
            f"Input action on a select element, please try to use select action on this element. element_id={element_id}"
        )


class FailToClick(SkyvernException):
    def __init__(self, element_id: str, anchor: str = "self"):
        super().__init__(f"Failed to click({anchor}). element_id={element_id}")


class FailToSelectByLabel(SkyvernException):
    def __init__(self, element_id: str):
        super().__init__(f"Failed to select by label. element_id={element_id}")


class FailToSelectByIndex(SkyvernException):
    def __init__(self, element_id: str):
        super().__init__(f"Failed to select by index. element_id={element_id}")


class OptionIndexOutOfBound(SkyvernException):
    def __init__(self, element_id: str):
        super().__init__(f"Option index is out of bound. element_id={element_id}")


class FailToSelectByValue(SkyvernException):
    def __init__(self, element_id: str):
        super().__init__(f"Failed to select by value. element_id={element_id}")


class EmptySelect(SkyvernException):
    def __init__(self, element_id: str):
        super().__init__(
            f"nothing is selected, try to select again. element_id={element_id}",
        )


class TaskAlreadyCanceled(SkyvernHTTPException):
    def __init__(self, new_status: str, task_id: str):
        super().__init__(
            f"Invalid task status transition to {new_status} for {task_id} because task is already canceled"
        )


class InvalidTaskStatusTransition(SkyvernHTTPException):
    def __init__(self, old_status: str, new_status: str, task_id: str):
        super().__init__(f"Invalid task status transition from {old_status} to {new_status} for {task_id}")


class ErrFoundSelectableElement(SkyvernException):
    def __init__(self, element_id: str, err: Exception):
        super().__init__(
            f"error when selecting elements in the children list. element_id={element_id}, error={repr(err)}"
        )


class NoSelectableElementFound(SkyvernException):
    def __init__(self, element_id: str):
        super().__init__(f"No selectable elements found in the children list. element_id={element_id}")


class HttpException(SkyvernException):
    def __init__(self, status_code: int, url: str, msg: str | None = None) -> None:
        super().__init__(f"HTTP Exception, status_code={status_code}, url={url}" + (f", msg={msg}" if msg else ""))


class WrongElementToUploadFile(SkyvernException):
    def __init__(self, element_id: str):
        super().__init__(
            f"No file chooser dialog opens, so file can't be uploaded through element {element_id}. Please try to upload again with another element."
        )


class FailedToFetchSecret(SkyvernException):
    def __init__(self) -> None:
        super().__init__("Failed to get the actual value of the secret parameter")


class NoIncrementalElementFoundForCustomSelection(SkyvernException):
    def __init__(self, element_id: str) -> None:
        super().__init__(
            f"No incremental element found, maybe try an input action or taking the select action on other elements. element_id={element_id}"
        )


class NoAvailableOptionFoundForCustomSelection(SkyvernException):
    def __init__(self, reason: str | None) -> None:
        super().__init__(f"No available option to select. reason: {reason}.")


class NoElementMatchedForTargetOption(SkyvernException):
    def __init__(self, target: str, reason: str | None) -> None:
        super().__init__(
            f"No element matches for the target value, try another value. reason: {reason}.  target_value='{target}'."
        )


class NoElementBoudingBox(SkyvernException):
    def __init__(self, element_id: str) -> None:
        super().__init__(f"Element does not have a bounding box. element_id={element_id}")


class NoIncrementalElementFoundForAutoCompletion(SkyvernException):
    def __init__(self, element_id: str, text: str) -> None:
        super().__init__(f"No auto completion shown up after fill in [{text}]. element_id={element_id}")


class NoSuitableAutoCompleteOption(SkyvernException):
    def __init__(self, reasoning: str | None, target_value: str) -> None:
        super().__init__(
            f"No suitable auto complete option to choose. target_value={target_value}, reasoning={reasoning}"
        )


class NoAutoCompleteOptionMeetCondition(SkyvernException):
    def __init__(
        self, reasoning: str | None, required_relevance: float, target_value: str, closest_relevance: float
    ) -> None:
        super().__init__(
            f"No auto complete option meet the condition(relevance_float>{required_relevance}). reasoning={reasoning}, target_value={target_value}, closest_relevance={closest_relevance}"
        )


class ErrEmptyTweakValue(SkyvernException):
    def __init__(self, reasoning: str | None, current_value: str) -> None:
        super().__init__(
            f"Empty tweaked value for the current value. reasoning={reasoning}, current_value={current_value}"
        )


class FailToFindAutocompleteOption(SkyvernException):
    def __init__(self, current_value: str) -> None:
        super().__init__(
            f"Can't find a suitable auto completion for the current value, maybe retry with another reasonable value. current_value={current_value}"
        )


class IllegitComplete(SkyvernException):
    def __init__(self, data: dict | None = None) -> None:
        data_str = f", data={data}" if data else ""
        super().__init__(f"Illegit complete{data_str}")


class CachedActionPlanError(SkyvernException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidUrl(SkyvernHTTPException):
    def __init__(self, url: str) -> None:
        super().__init__(
            f"Invalid URL: {url}. Skyvern supports HTTP and HTTPS urls.", status_code=status.HTTP_400_BAD_REQUEST
        )


class BlockedHost(SkyvernHTTPException):
    def __init__(self, host: str) -> None:
        super().__init__(
            f"The host in your url is blocked: {host}",
            status_code=status.HTTP_400_BAD_REQUEST,
        )


class InvalidWorkflowParameter(SkyvernHTTPException):
    def __init__(self, expected_parameter_type: str, value: str, workflow_permanent_id: str | None = None) -> None:
        message = f"Invalid workflow parameter. Expected parameter type: {expected_parameter_type}. Value: {value}."
        if workflow_permanent_id:
            message += f" Workflow permanent id: {workflow_permanent_id}"
        super().__init__(
            message,
            status_code=status.HTTP_400_BAD_REQUEST,
        )


class InteractWithDisabledElement(SkyvernException):
    def __init__(self, element_id: str):
        super().__init__(
            f"The element(id={element_id}) now is disabled, try to interact with it later when it's enabled."
        )



from .utils.llms import call_model
from .utils.views import print_agent_output

TEMPLATE = """You are an expert research article reviewer. \
Your goal is to review research drafts and provide feedback to the reviser only based on specific guidelines. \
"""

class ReviewerAgent:
    def __init__(self, headers=None):
        self.headers = headers or {}

    async def review_draft(self, draft_state: dict):
        """
        Review a draft article
        :param draft_state:
        :return:
        """
        task = draft_state.get("task")
        guidelines = "- ".join(guideline for guideline in task.get("guidelines"))
        revision_notes = draft_state.get("revision_notes")

        revise_prompt = f"""The reviser has already revised the draft based on your previous review notes with the following feedback:
{revision_notes}\n
Please provide additional feedback ONLY if critical since the reviser has already made changes based on your previous feedback.
If you think the article is sufficient or that non critical revisions are required, please aim to return None.
"""

        review_prompt = f"""You have been tasked with reviewing the draft which was written by a non-expert based on specific guidelines.
Please accept the draft if it is good enough to publish, or send it for revision, along with your notes to guide the revision.
If not all of the guideline criteria are met, you should send appropriate revision notes.
If the draft meets all the guidelines, please return None.
{revise_prompt if revision_notes else ""}

Guidelines: {guidelines}\nDraft: {draft_state.get("draft")}\n
"""
        prompt = [{
            "role": "system",
            "content": TEMPLATE
        }, {
            "role": "user",
            "content": review_prompt
        }]

        response = call_model(prompt, model=task.get("model"), api_key=self.headers.get("openai_api_key"))

        if task.get("verbose"):
            if self.websocket and self.stream_output:
                await self.stream_output("logs", "review_feedback", f"Review feedback is: {response}...", self.websocket)
            else:
                print_agent_output(f"Review feedback is: {response}...", agent="REVIEWER")

        if "None" in response:
            return None
        return response

    def run(self, draft_state: dict):
        task = draft_state.get("task")
        guidelines = task.get("guidelines")
        to_follow_guidelines = task.get("follow_guidelines")
        review = None
        if to_follow_guidelines:
            print_agent_output("Reviewing draft...", agent="REVIEWER")

            if task.get("verbose"):
                print_agent_output(f"Following guidelines {guidelines}...", agent="REVIEWER")

            review = self.review_draft(draft_state)
        else:
            print_agent_output("Ignoring guidelines...", agent="REVIEWER")
        return {"review": review}



import asyncio
import logging
import traceback
from datetime import datetime, timezone
from typing import Any, Callable, Dict, Optional, Union
from uuid import UUID

from autogen_agentchat.base._task import TaskResult
from autogen_agentchat.messages import (
    AgentEvent,
    ChatMessage,
    HandoffMessage,
    MultiModalMessage,
    StopMessage,
    TextMessage,
    ToolCallExecutionEvent,
    ToolCallRequestEvent,
)
from autogen_core import CancellationToken
from fastapi import WebSocket, WebSocketDisconnect

from .database import DatabaseManager
from .models.ag_db import Message, MessageConfig, Run, RunStatus, TeamResult
from .teammanager import TeamManager

logger = logging.getLogger(__name__)


class WebSocketManager:
    """Manages WebSocket connections and message streaming for team task execution"""

    def __init__(self, db_manager: DatabaseManager):
        self.db_manager = db_manager
        self._connections: Dict[UUID, WebSocket] = {}
        self._cancellation_tokens: Dict[UUID, CancellationToken] = {}
        # Track explicitly closed connections
        self._closed_connections: set[UUID] = set()
        self._input_responses: Dict[UUID, asyncio.Queue] = {}

        self._cancel_message = TeamResult(
            task_result=TaskResult(
                messages=[TextMessage(source="user", content="Run cancelled by user")],
                stop_reason="cancelled by user",
            ),
            usage="",
            duration=0,
        ).model_dump()

    def _get_stop_message(self, reason: str) -> dict:
        return TeamResult(
            task_result=TaskResult(
                messages=[TextMessage(source="user", content=reason)],
                stop_reason=reason,
            ),
            usage="",
            duration=0,
        ).model_dump()

    async def connect(self, websocket: WebSocket, run_id: UUID) -> bool:
        try:
            await websocket.accept()
            self._connections[run_id] = websocket
            self._closed_connections.discard(run_id)
            # Initialize input queue for this connection
            self._input_responses[run_id] = asyncio.Queue()

            await self._send_message(
                run_id,
                {
                    "type": "system",
                    "status": "connected",
                    "timestamp": datetime.now(timezone.utc).isoformat(),
                },
            )

            return True
        except Exception as e:
            logger.error(f"Connection error for run {run_id}: {e}")
            return False

    async def start_stream(self, run_id: UUID, task: str, team_config: dict) -> None:
        """Start streaming task execution with proper run management"""
        if run_id not in self._connections or run_id in self._closed_connections:
            raise ValueError(f"No active connection for run {run_id}")

        team_manager = TeamManager()
        cancellation_token = CancellationToken()
        self._cancellation_tokens[run_id] = cancellation_token
        final_result = None

        try:
            # Update run with task and status
            run = await self._get_run(run_id)
            if run:
                run.task = MessageConfig(content=task, source="user").model_dump()
                run.status = RunStatus.ACTIVE
                self.db_manager.upsert(run)

            input_func = self.create_input_func(run_id)

            async for message in team_manager.run_stream(
                task=task,
                team_config=team_config,
                input_func=input_func,
                cancellation_token=cancellation_token,
            ):
                if (
                    cancellation_token.is_cancelled()
                    or run_id in self._closed_connections
                ):
                    logger.info(
                        f"Stream cancelled or connection closed for run {run_id}"
                    )
                    break

                formatted_message = self._format_message(message)
                if formatted_message:
                    await self._send_message(run_id, formatted_message)

                    # Save messages by concrete type
                    if isinstance(
                        message,
                        (
                            TextMessage,
                            MultiModalMessage,
                            StopMessage,
                            HandoffMessage,
                            ToolCallRequestEvent,
                            ToolCallExecutionEvent,
                        ),
                    ):
                        await self._save_message(run_id, message)
                    # Capture final result if it's a TeamResult
                    elif isinstance(message, TeamResult):
                        final_result = message.model_dump()
            if (
                not cancellation_token.is_cancelled()
                and run_id not in self._closed_connections
            ):
                if final_result:
                    await self._update_run(
                        run_id, RunStatus.COMPLETE, team_result=final_result
                    )
                else:
                    logger.warning(
                        f"No final result captured for completed run {run_id}"
                    )
                    await self._update_run_status(run_id, RunStatus.COMPLETE)
            else:
                await self._send_message(
                    run_id,
                    {
                        "type": "completion",
                        "status": "cancelled",
                        "data": self._cancel_message,
                        "timestamp": datetime.now(timezone.utc).isoformat(),
                    },
                )
                # Update run with cancellation result
                await self._update_run(
                    run_id, RunStatus.STOPPED, team_result=self._cancel_message
                )

        except Exception as e:
            logger.error(f"Stream error for run {run_id}: {e}")
            traceback.print_exc()
            await self._handle_stream_error(run_id, e)
        finally:
            self._cancellation_tokens.pop(run_id, None)

    async def _save_message(
        self, run_id: UUID, message: Union[AgentEvent | ChatMessage, ChatMessage]
    ) -> None:
        """Save a message to the database"""

        run = await self._get_run(run_id)
        if run:
            db_message = Message(
                session_id=run.session_id,
                run_id=run_id,
                config=message.model_dump(),
                user_id=None,  # You might want to pass this from somewhere
            )
            self.db_manager.upsert(db_message)

    async def _update_run(
        self,
        run_id: UUID,
        status: RunStatus,
        team_result: Optional[dict] = None,
        error: Optional[str] = None,
    ) -> None:
        """Update run status and result"""
        run = await self._get_run(run_id)
        if run:
            run.status = status
            if team_result:
                run.team_result = team_result
            if error:
                run.error_message = error
            self.db_manager.upsert(run)

    def create_input_func(self, run_id: UUID) -> Callable:
        """Creates an input function for a specific run"""

        async def input_handler(
            prompt: str = "", cancellation_token: Optional[CancellationToken] = None
        ) -> str:
            try:
                # Send input request to client
                await self._send_message(
                    run_id,
                    {
                        "type": "input_request",
                        "prompt": prompt,
                        "data": {"source": "system", "content": prompt},
                        "timestamp": datetime.now(timezone.utc).isoformat(),
                    },
                )

                # Wait for response
                if run_id in self._input_responses:
                    response = await self._input_responses[run_id].get()
                    return response
                else:
                    raise ValueError(f"No input queue for run {run_id}")

            except Exception as e:
                logger.error(f"Error handling input for run {run_id}: {e}")
                raise

        return input_handler

    async def handle_input_response(self, run_id: UUID, response: str) -> None:
        """Handle input response from client"""
        if run_id in self._input_responses:
            await self._input_responses[run_id].put(response)
        else:
            logger.warning(f"Received input response for inactive run {run_id}")

    async def stop_run(self, run_id: UUID, reason: str) -> None:
        if run_id in self._cancellation_tokens:
            logger.info(f"Stopping run {run_id}")

            stop_message = self._get_stop_message(reason)

            try:
                # Update run record first
                await self._update_run(
                    run_id, status=RunStatus.STOPPED, team_result=stop_message
                )

                # Then handle websocket communication if connection is active
                if (
                    run_id in self._connections
                    and run_id not in self._closed_connections
                ):
                    await self._send_message(
                        run_id,
                        {
                            "type": "completion",
                            "status": "cancelled",
                            "data": stop_message,
                            "timestamp": datetime.now(timezone.utc).isoformat(),
                        },
                    )

                # Finally cancel the token
                self._cancellation_tokens[run_id].cancel()

            except Exception as e:
                logger.error(f"Error stopping run {run_id}: {e}")
                # We might want to force disconnect here if db update failed
                # await self.disconnect(run_id)  # Optional

    async def disconnect(self, run_id: UUID) -> None:
        """Clean up connection and associated resources"""
        logger.info(f"Disconnecting run {run_id}")

        # Mark as closed before cleanup to prevent any new messages
        self._closed_connections.add(run_id)

        # Cancel any running tasks
        await self.stop_run(run_id, "Connection closed")

        # Clean up resources
        self._connections.pop(run_id, None)
        self._cancellation_tokens.pop(run_id, None)
        self._input_responses.pop(run_id, None)

    async def _send_message(self, run_id: UUID, message: dict) -> None:
        """Send a message through the WebSocket with connection state checking

        Args:
            run_id: UUID of the run
            message: Message dictionary to send
        """
        if run_id in self._closed_connections:
            logger.warning(
                f"Attempted to send message to closed connection for run {run_id}"
            )
            return

        try:
            if run_id in self._connections:
                websocket = self._connections[run_id]
                await websocket.send_json(message)
        except WebSocketDisconnect:
            logger.warning(
                f"WebSocket disconnected while sending message for run {run_id}"
            )
            await self.disconnect(run_id)
        except Exception as e:
            logger.error(f"Error sending message for run {run_id}: {e}, {message}")
            # Don't try to send error message here to avoid potential recursive loop
            await self._update_run_status(run_id, RunStatus.ERROR, str(e))
            await self.disconnect(run_id)

    async def _handle_stream_error(self, run_id: UUID, error: Exception) -> None:
        """Handle stream errors with proper run updates"""
        if run_id not in self._closed_connections:
            error_result = TeamResult(
                task_result=TaskResult(
                    messages=[TextMessage(source="system", content=str(error))],
                    stop_reason="An error occurred while processing this run",
                ),
                usage="",
                duration=0,
            ).model_dump()

            await self._send_message(
                run_id,
                {
                    "type": "completion",
                    "status": "error",
                    "data": error_result,
                    "timestamp": datetime.now(timezone.utc).isoformat(),
                },
            )

            await self._update_run(
                run_id, RunStatus.ERROR, team_result=error_result, error=str(error)
            )

    def _format_message(self, message: Any) -> Optional[dict]:
        """Format message for WebSocket transmission

        Args:
            message: Message to format

        Returns:
            Optional[dict]: Formatted message or None if formatting fails
        """

        try:
            if isinstance(message, MultiModalMessage):
                message_dump = message.model_dump()
                message_dump["content"] = [
                    message_dump["content"][0],
                    {
                        "url": f"data:image/png;base64,{message_dump['content'][1]['data']}",
                        "alt": "WebSurfer Screenshot",
                    },
                ]
                return {"type": "message", "data": message_dump}

            elif isinstance(message, TeamResult):
                return {
                    "type": "result",
                    "data": message.model_dump(),
                    "status": "complete",
                }

            elif isinstance(
                message,
                (
                    TextMessage,
                    StopMessage,
                    HandoffMessage,
                    ToolCallRequestEvent,
                    ToolCallExecutionEvent,
                ),
            ):
                return {"type": "message", "data": message.model_dump()}

            return None

        except Exception as e:
            logger.error(f"Message formatting error: {e}")
            return None

    async def _get_run(self, run_id: UUID) -> Optional[Run]:
        """Get run from database

        Args:
            run_id: UUID of the run to retrieve

        Returns:
            Optional[Run]: Run object if found, None otherwise
        """
        response = self.db_manager.get(Run, filters={"id": run_id}, return_json=False)
        return response.data[0] if response.status and response.data else None

    async def _update_run_status(
        self, run_id: UUID, status: RunStatus, error: Optional[str] = None
    ) -> None:
        """Update run status in database

        Args:
            run_id: UUID of the run to update
            status: New status to set
            error: Optional error message
        """
        run = await self._get_run(run_id)
        if run:
            run.status = status
            run.error_message = error
            self.db_manager.upsert(run)

    async def cleanup(self) -> None:
        """Clean up all active connections and resources when server is shutting down"""
        logger.info(f"Cleaning up {len(self.active_connections)} active connections")

        try:
            # First cancel all running tasks
            for run_id in self.active_runs.copy():
                if run_id in self._cancellation_tokens:
                    self._cancellation_tokens[run_id].cancel()
                run = await self._get_run(run_id)
                if run and run.status == RunStatus.ACTIVE:
                    interrupted_result = TeamResult(
                        task_result=TaskResult(
                            messages=[
                                TextMessage(
                                    source="system",
                                    content="Run interrupted by server shutdown",
                                )
                            ],
                            stop_reason="server_shutdown",
                        ),
                        usage="",
                        duration=0,
                    ).model_dump()

                    run.status = RunStatus.STOPPED
                    run.team_result = interrupted_result
                    self.db_manager.upsert(run)

            # Then disconnect all websockets with timeout
            # 10 second timeout for entire cleanup
            async with asyncio.timeout(10):
                for run_id in self.active_connections.copy():
                    try:
                        # Give each disconnect operation 2 seconds
                        async with asyncio.timeout(2):
                            await self.disconnect(run_id)
                    except asyncio.TimeoutError:
                        logger.warning(f"Timeout disconnecting run {run_id}")
                    except Exception as e:
                        logger.error(f"Error disconnecting run {run_id}: {e}")

        except asyncio.TimeoutError:
            logger.warning("WebSocketManager cleanup timed out")
        except Exception as e:
            logger.error(f"Error during WebSocketManager cleanup: {e}")
        finally:
            # Always clear internal state, even if cleanup had errors
            self._connections.clear()
            self._cancellation_tokens.clear()
            self._closed_connections.clear()
            self._input_responses.clear()

    @property
    def active_connections(self) -> set[UUID]:
        """Get set of active run IDs"""
        return set(self._connections.keys()) - self._closed_connections

    @property
    def active_runs(self) -> set[UUID]:
        """Get set of runs with active cancellation tokens"""
        return set(self._cancellation_tokens.keys())






"""
仅作参考，暂时未用上


"""














base_system_message = """PersonalAssistant:
===CONSTRAINTS===
If you are missing any information or details to complete a task, you ask for clarification.
You think step by step to make sure you have the correct solution

===RESPONSE FORMAT[STRICT - MARKDOWN]===
ALWAYS add a new line after ``` in markdown for my GUI to render it correctly. Never add language to the codeblock.
Example:
```


print('Hello World')

```
"""

function_res_agent = """Function Response Agent:
===CONSTRAINTS===
Your job is to format the responses according to the response format section
You always follow the response format section
Write all responses as MARKDOWN
ALWAYS add a new line after codeblocks in markdown for my GUI to render it correctly.
Example:
```python

print('Hello World')

```

===RESPONSE FORMAT[STRICT]===
- If Results is empty, return "No results found". Do not invent results.
- If successful:
**wikidata_sparql_query**:
Return response in human readable format
FORMAT:
```markdown
### [TOPIC]
For each Entry:
#### [Label](if wikipedia_article else wikidata_entry)
- query results
```
**scrape_webpage**:
Return the full text content of the webpage (unless user has specified a summary/abstract).
ALWAYS return the code examples from the webpage
**write_file**:
Return the filename of the saved file.
Do NOT the content of the file
**write_code**:
Return the filename of the saved file.
Do NOT return the content of the file
**edit_file**:
Return the filename of the saved file.
Return the changes made to the file
Do NOT return the other content of the file
**knowledgebase_create_entry**:
Return the filename of the saved file.
Do NOT the content of the file
**write_history_entry**:
Return the filename of the saved file.
Do NOT return the content of the file
**python_repl**:
If the code saves a file, return the filename of the saved file.
If the code does not save a file, return the output of the code
If the output is empty/the code runs a process, return "Code ran successfully"
Do NOT return the code
**image_to_text**:
Return the text caption/description
"""

review_agent = """Review Agent:
===CONSTRAINTS===
You are a Code Reviwer.
You follow the review format laid out in the response format section.
You are an incredibly high level programmer in every language.
You review honestly. Good code comes before everything.

===RESPONSE FORMAT[STRICT]===
- Write all responses as MARKDOWN

Review:
- Error-handling Suggestions;
- Performance Suggestions;
- Best-practice Suggestions;
- Security Suggestions;

ALWAYS add a new line after ```language in markdown for my GUI to render it correctly
Example:
```python

print('Hello World')

```
"""

brainstorm_agent = """Brainstorm Agent:
===CONSTRAINTS===
You follow the brainstorm format laid out in the response format section.
Emulate the thinking and speaking style of an experienced senior software engineer
You think step by step to make sure you have the most logically sound brainstorming ideas. Good ideas come before everything.
Do NOT write code examples. Purely higher level ideas.

===RESPONSE FORMAT[STRICT]===
- Write all responses as MARKDOWN

Brainstorm:
- Problem;
- Approach;
- Technology;

ALWAYS add a new line after ```language in markdown for my GUI to render it correctly
Example:
```python

print('Hello World')

```
"""

spec_writer = """Specification Writer:
===CONSTRAINTS===
Emulate the thinking and speaking style of an experienced senior software engineer
Your role is to provide a clear and detailed specification following the receipt of a ticket.
This specification should serve as a comprehensive guide for a senior developer.
Approach the task methodically to ensure clarity, accuracy, and thoroughness, just as an experienced senior software engineer would.

===RESPONSE FORMAT[STRICT]===

Write all responses as MARKDOWN

Specification:
- Title;
- Description;
- Desired Functionality;
- Solution Structure;
- User Stories;
- Unit Tests Skeletons;
- Acceptance Criteria;

ALWAYS add a new line after ```language in markdown for the GUI to render correctly.
Example:
```python

print('Hello World')

```
"""

ticket_agent = """Ticket Agent:
===CONSTRAINTS===
You follow the ticket format laid out in the response format section.
Emulate the thinking and speaking style of an experienced senior software engineer
You think step by step to make sure you have the most logically sound ticket. Good tickets come before everything.

===RESPONSE FORMAT[STRICT]===
- Write all responses as MARKDOWN

Ticket:
- Title;
- Description;
- Technology;
- Requirements;
- File Structure;
- File Skeletons;
- Acceptance Criteria;

ALWAYS add a new line after ```language in markdown for my GUI to render it correctly
Example:
```python

print('Hello World')

```
"""

unit_test_writer = """Unit Test Writer:
===CONSTRAINTS===
Emulate the thinking and speaking style of an experienced senior software engineer
Your role is to provide clear and detailed unit tests as specified in the Ticket to enable Test Driven Development.
Approach the task methodically to ensure clarity, accuracy, and thoroughness, just as an experienced senior software engineer would.
Think step by step to be sure you have the most logically sound unit tests. Good unit tests come before everything.
Only write Code. Any text should be in comments.
"""


code_writer = """Code Writer:
===CONSTRAINTS===
Only write Code. Any text should be in comments.
Emulate the thinking and speaking style of an experienced senior software engineer.
Your role is to translate the provided Ticket or file from Ticket into clear, logical, and concise code.
This process should be methodically undertaken to ensure that all features and requirements are correctly implemented.
Methodically approach the task to ensure clarity, accuracy, and thoroughness.
"""

suggest_changes_agent = """
Suggest Changes Agent:

===CONSTRAINTS===
As if you were an experienced senior software engineer, approach the task in a systematic and methodical manner, ensuring clarity, accuracy, and comprehensiveness in your suggestions.
Your suggestions should specify the exact range of lines to change, and describe in detail what changes should be made.
Every line from the start_line to the end_line will be replaced by the new code, therefore it's crucial to be precise in your line numbering.
Avoid suggesting changes to content that isn't going to be replaced.

===RESPONSE FORMAT[STRICT]===
The file number range should always be provided as [start_line, end_line], even if the change only affects a single line.
When providing your suggestions, ensure you specify the exact lines that need to be changed. Mistakes in line references can lead to confusion or incorrect changes.
Detailed explanations of the changes and clear reasoning behind them are also required to facilitate understanding of your suggestions.
"""

help_agent = """### Available commands:
* /python - python interpreter
* /wikidata - wikidata sparql handler
* /review - performs a review of code following a strict response format
* /brainstorm [n] - returns a list of n ideas for the topic following a strict response format
* /write_spec - Writes a spec for the brainstorm n following a strict response format
* /ticket - creates a ticket for the brainstorm/spec following a strict response format
* /write_tests - writes tests for the ticket / file from ticket
* /write_code - writes code for the ticket / file from ticket
"""



import logging
import os
import re
from pathlib import Path

from mtmai.core import coreutils
from mtmai.core.config import settings
from mtmai.mtlibs import mtutils
from mtmai.mtlibs.mtutils import command_exists, is_in_gitpod, npm_patch_version
from mtmai.mtlibs.temboio import run_tmpbo_instance1
from mtmlib.github import git_commit_push
from mtmlib.mtutils import bash

logger = logging.getLogger()


async def init_project():
    is_in_docker_build = os.getenv("X_DOCKER_BUILD")
    if is_in_docker_build:
        pass
    if not coreutils.is_in_gitpod():
        return
    docker_config = Path.home().joinpath(".docker/config.json")
    # if settings.DOCKERHUB_PASSWORD and not docker_config.exists():
    #     bash(
    #         f"(command -v docker && echo {settings.DOCKERHUB_PASSWORD} | docker login --username {
    #             settings.DOCKERHUB_USER} --password-stdin) || true"
    #     )
    # if settings.NPM_TOKEN:
    #     if is_in_gitpod():
    #         Path.home().joinpath(".npmrc").write_text(
    #             f"//registry.npmjs.org/:_authToken={settings.NPM_TOKEN}\n"
    #         )

    # from mtmlib import tunnel
    # threading.Thread(target=lambda: asyncio.run(
    #     tunnel.start_cloudflared())).start()
    # from mtmai.mtlibs.server.kasmvnc import run_kasmvnc
    # threading.Thread(target=run_kasmvnc).start()
    # 拉取第三方项目
    # await git_clone_repos(
    #     base_dir=settings.storage_dir,
    #     repo_urls=["https://github.com/Skyvern-AI/skyvern"],
    # )
    # bash("command -v playwright || ( python -m pip install playwright)")
    # bash("playwright install-deps")
    # bash("playwright install")
    # bash("vnc")


def hf_trans1_commit():
    target_dir = (
        Path(settings.storage_dir)
        .joinpath(settings.gitsrc_dir)
        .joinpath(settings.HUGGINGFACEHUB_DEFAULT_WORKSPACE)
    )
    rnd_str = mtutils.gen_orm_id_key()
    Path(target_dir).joinpath("Dockerfile").write_text(f"""
# {rnd_str}
FROM docker.io/gitgit188/tmpboai
ENV DATABASE_URL={settings.MTMAI_DATABASE_URL}
ENV LOKI_USER={settings.LOKI_USER}
ENV GRAFANA_TOKEN={settings.GRAFANA_TOKEN}
ENV LOKI_ENDPOINT={settings.LOKI_ENDPOINT}
RUN sudo apt update

""")
    Path(target_dir).joinpath("README.md").write_text(f"""---
title: Trans1
emoji: 🏢
colorFrom: red
colorTo: gray
sdk: docker
pinned: false
license: other
app_port:  {settings.FRONT_PORT}
---""")
    bash(f"cd {target_dir} && git commit -am abccommit && git push")
    return {"ok": True}


def run_clean():
    bun_cache_dir = Path.home().joinpath(".bun/install/cache")
    bash(f"rm -rdf {bun_cache_dir}")

    if command_exists("pip"):
        logging.info("正在清理 pip 缓存")
        bash("pip cache dir && pip cache purge")
    if command_exists("docker"):
        logging.info("正在清理 docker 缓存")
        bash("docker system prune -f")

    if command_exists("pyenv"):
        bash("pyenv rehash")  # 可能不正确
    if is_in_gitpod():
        logger.info("删除 ~/.rustup")
        bash("rm -rdf ~/.rustup")
        logger.info("删除 ~/.rvm")
        dotrvm = Path.home().joinpath(".rvm")
        if dotrvm.exists():
            bash("rm -rdf ~/.rvm")


def docker_build_base():
    logger.info("🚀 build docker image_base")
    image_tag = f"{settings.DOCKERHUB_USER}/base"
    bash(
        f"docker build --progress=plain -t {
            image_tag} -f Dockerfile.base . && docker push {image_tag}"
    )
    logger.info("✅ build docker image_base")


async def run_deploy():
    await dp_vercel_mtmaiadmin()
    # await dp_vercel_mtmaifront()
    await run_tmpbo_instance1()
    logger.info("✅ tembo io pushed")
    hf_trans1_commit()
    logger.info("✅ hf_space_commit")
    git_commit_push()
    logger.info("✅ git_commit_push")


py_projects = [
    "mtmai",
    "mtmlib",
    "mtmtrain",
    "mtmaisdk",
]


def run_testing():
    for project in py_projects:
        if Path(f"{project}/{project}/tests").exists():
            bash(f"cd {project} && coverage run -m pytest ")
            logger.info("✅ testing ok!")


def release_py():
    # version_tag = read_tag()
    # logger.info("version tag: %s", version_tag)
    # run_testing()
    # for project in py_projects:
    #     dist_dir = Path(f"{project}/dist")
    #     if dist_dir.exists():
    #         bash(f"rm -rdf {dist_dir}")
    #     bash(f"cd pyprojects/{project} && poetry build")

    # for project in py_projects:
    #     try:
    #         bash(f"cd pyprojects/{project} && poetry publish")
    #     except Exception as e:  # noqa: BLE001
    #         logger.info("⚠ pypi %s 发布失败 %s", project, e)

    for project in py_projects:
        py_project_dir = f"pyprojects/{project}"
        mtutils.pyproject_patch_version(Path(py_project_dir).joinpath("pyproject.toml"))

    release_npm()
    # next_version = patch_git_tag_version()
    # logger.info("✅ patch_git_tag_version ok!,next version tag: %s", next_version)


def release_npm():
    npm_packages = [
        # "apps/mtmaiweb",
        "apps/mtmaiadmin",
        # "apps/mtmaifront",
        # "packages/mtmscreentocode",
        "packages/mtxuilib",
        "packages/mtmaiapi",
    ]

    bash("bun run turbo build")

    for package in npm_packages:
        npm_patch_version(package)
    NPM_TOKEN = os.getenv("NPM_TOKEN")
    print("NPM_TOKEN", NPM_TOKEN)
    if NPM_TOKEN:
        bash(
            f"export NPM_TOKEN={NPM_TOKEN} && bun run changeset publish --token {NPM_TOKEN} --no-git-tag"
        )
    else:
        raise Exception("NPM_TOKEN 未设置")

    logger.info("✅ release_npm ok!")


async def dp_cfpage():
    logger.info("🚀 正在部署 cfpage")
    from mtmlib import vercel

    pages = ["apps/mtmaiweb/src/app/(web)/layout.tsx"]
    pre_content = Path(pages[0]).read_text()
    # Check if the current runtime is nodejs before replacing
    if re.search(r'runtime\s*=\s*["\']nodejs["\']', pre_content):
        new_content = re.sub(r'runtime\s*=\s*"nodejs"', r'runtime="edge"', pre_content)
        Path(pages[0]).write_text(new_content)
        logger.info("✅ 替换layout 源码中的 runtime 为 edge")
    else:
        new_content = pre_content

    vercel.deploy_vercel(
        project_dir="apps/mtmaiweb",
        is_cfpage=True,
        project_name="mtmaiweb",
        vercel_token=settings.vercel_token,
    )

    # 恢复
    Path(pages[0]).write_text(pre_content)
    logger.info("✅ 恢复layout 源码中的 runtime 为原来的值")
    git_commit_push()


async def dp_vercel_mtmaiadmin():
    logger.info("🚀 正在部署到 vercel")
    from mtmlib import vercel

    vercel.deploy_vercel(
        project_dir="apps/mtmaiadmin",
        is_cfpage=False,
        project_name="mtmaiadmin",
        vercel_token=settings.vercel_token,
        deploy_to_vercel=True,
        build_local=False,
    )
    git_commit_push()


async def dp_vercel_mtmaifront():
    logger.info("🚀 正在部署到 vercel")
    from mtmlib import vercel

    vercel.deploy_vercel(
        project_dir="apps/mtmaifront",
        is_cfpage=False,
        project_name="mtmaifront",
        vercel_token=settings.vercel_token,
        deploy_to_vercel=True,
        build_local=False,
    )
    git_commit_push()



from typing import Any, TypeVar

from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


def is_basemodel_subclass(model: Any) -> bool:
    try:
        return issubclass(model, BaseModel)
    except TypeError:
        return False



from enum import member
from textwrap import dedent
from typing import Literal

from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables import Runnable, RunnableConfig
from langgraph.prebuilt import tools_condition
from pydantic import BaseModel

from mtmai.models.graph_config import HomeChatState, MainState



def edge_supervisor(state: MainState):
    is_tools = tools_condition(state)
    if is_tools == "tools":
        return "chat_tools_node"
    next_to = state.get("next")
    if next_to:
        return next_to
    return "__end__"


members = ["human_chat", "JokeWriter"]

class RouteResponse(BaseModel):
    next: Literal["human_chat", "JokeWriter"]

# supervisor_prompt = dedent("""
# Given the conversation above, who should act next?
# Or should we FINISH? Select one of: {options}
# """)


class SupervisorNode:
    def __init__(self, runnable: Runnable):
        self.runnable = runnable

    def agent_name(self):
        return "Supervisor"

    async def __call__(self, state: HomeChatState, config: RunnableConfig):
        options = member
        supervisor_system_prompt = (
            "You are a supervisor tasked with managing a conversation between the"
            " following workers:  {members}. \n"
            "Given the following user request,"
            " respond with the worker to act next. Each worker will perform a"
            " task and respond with their results and status. When finished,"
            " respond with FINISH."
        )

        options = [*members]

        prompt = ChatPromptTemplate.from_messages(
            [
                ("system", supervisor_system_prompt),
                MessagesPlaceholder(variable_name="messages", optional=True),
                (
                    "system",
                    "Given the conversation above, who should act next?"
            " Or should we FINISH? Select one of: {options}",
                ),
            ]
        ).partial(options=str(options), members=", ".join(members))

        supervisor_chain = prompt | self.runnable.with_structured_output(RouteResponse)
        result = await supervisor_chain.ainvoke(state.model_dump())
        return {
            "next": result.next
        }



import json
import logging

import requests
from crewai import Agent, Crew, Task
from langchain.tools import tool
from unstructured.partition.html import partition_html

# from mtmai.mtlibs.aiutils import lcllm_openai_chat
from mtmai.llm.llm import get_llm_chatbot_default

logger = logging.getLogger()


class BrowserTools:
    @tool("Scrape website content")
    async def scrape_and_summarize_website(website):
        """Useful to scrape and summarize a website content"""
        # url = f"https://chrome.browserless.io/content?token={os.environ['BROWSERLESS_API_KEY']}"
        url = "http://172.17.0.1:13001"

        logger.info('调用工具 BrowserTools.scrape_and_summarize_website: "%s"', website)

        payload = json.dumps({"url": website})
        headers = {"cache-control": "no-cache", "content-type": "application/json"}
        response = requests.request(
            "POST", url, headers=headers, data=payload, timeout=30
        )
        elements = partition_html(text=response.text)
        content = "\n\n".join([str(el) for el in elements])
        content = [content[i : i + 8000] for i in range(0, len(content), 8000)]
        summaries = []

        llm = get_llm_chatbot_default()
        for chunk in content:
            agent = Agent(
                role="Principal Researcher",
                goal="Do amazing researches and summaries based on the content you are working with",
                backstory="You're a Principal Researcher at a big company and you need to do a research about a given topic.",
                allow_delegation=False,
                llm=llm,
            )
            task = Task(
                agent=agent,
                description=f"Analyze and summarize the content bellow, make sure to include the most relevant information in the summary, return only the summary nothing else.\n\nCONTENT\n----------\n{chunk}",
                expected_output="A bullet list summary of the top 5 most important AI news",  # 需要正确设置
            )
            # summary = task.execute()
            # Execute the crew
            crew = Crew(agents=[agent], tasks=[task], verbose=2)
            result = crew.kickoff()

            summaries.append(task.output.summary)
        return "\n\n".join(summaries)



from langchain_core.runnables import RunnableConfig

from mtmai.core.logging import get_logger
from mtmai.models.graph_config import HomeChatState
import structlog

LOG = structlog.get_logger()



class HumanInputNode:
    def __init__(self):
        pass

    async def __call__(self, state: HomeChatState, config: RunnableConfig):
        LOG.info("进入 HumanInputNode ")
        messages = state.messages
        user_input = state.user_input
        if user_input == "/1":
            LOG.info("特殊指令1")
            return {
                "next": "articleGen"
            }

        if user_input == "/2":
            LOG.info("特殊指令2")
            return {
                "next": "site"
            }
        return {
            "messages": messages,
            "next": "assistant",
        }



from datetime import datetime

from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import Runnable, RunnableConfig
from langgraph.graph import END, StateGraph
from langgraph.prebuilt import tools_condition

from mtmai.agents.ctx import mtmai_context
from mtmai.agents.graphutils import (
    CompleteOrEscalate,
    create_entry_node,
    create_tool_node_with_fallback,
)
from mtmai.core.logging import get_logger
from mtmai.models.graph_config import HomeChatState

logger = get_logger()

develop_safe_tools = []
develop_sensitive_tools = []
develop_tools = develop_safe_tools + develop_sensitive_tools


def route_develop_mode(
    state: HomeChatState,
):
    route = tools_condition(state.messages)
    if route == END:
        return END
    tool_calls = state.messages[-1].tool_calls
    did_cancel = any(tc["name"] == CompleteOrEscalate.__name__ for tc in tool_calls)
    if did_cancel:
        return "leave_skill"
    safe_toolnames = [t.name for t in develop_safe_tools]
    if all(tc["name"] in safe_toolnames for tc in tool_calls):
        return "develop_safe_tools"
    return "develop_sensitive_tools"


class DevelopNode:
    def __init__(self, runnable: Runnable):
        self.runnable = runnable

    def node_name(self):
        return "develop_node"

    async def __call__(self, state: HomeChatState, config: RunnableConfig):
        develop_prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    "You are a specialized assistant for handling development tasks. "
                    "The primary assistant delegates work to you whenever the user needs help with development-related queries or tasks. "
                    "You can use various tools to assist developers, such as reading logs, retrieving system status information, and performing other development-related actions. "
                    "Be thorough and precise in your responses, providing detailed information when necessary. "
                    "If a task requires multiple steps, guide the user through each step clearly. "
                    "If you need more information or if the task is beyond your capabilities, escalate the task back to the main assistant. "
                    "Remember to use the appropriate tools for each task and provide clear explanations of your actions. "
                    "Always prioritize security and best practices in your recommendations."
                    "\nCurrent time: {time}."
                    "\n\nIf the user needs help, and none of your tools are appropriate for it, then"
                    ' "CompleteOrEscalate" the dialog to the host assistant. Do not waste the user\'s time. Do not make up invalid tools or functions.'
                    "{additional_instructions}",
                ),
                ("placeholder", "{messages}"),
            ]
        ).partial(time=datetime.now())

        # develop_runnable = develop_prompt | self.runnable.bind_tools(
        #     develop_tools + [CompleteOrEscalate]
        # )
        return {
            "messages": await mtmai_context.ainvoke_model(
                develop_prompt, state, tools=develop_tools + [CompleteOrEscalate]
            )
        }

    @classmethod
    async def addto_primary_assistant(self, wf: StateGraph):
        llm_runnable = await mtmai_context.get_llm_openai("chat")

        wf.add_node(
            "enter_develop_mode",
            create_entry_node("Developer Mode & Development Assistant", "develop_mode"),
        )
        wf.add_node("develop_mode", DevelopNode(llm_runnable))
        wf.add_edge("enter_develop_mode", "develop_mode")
        wf.add_node(
            "develop_sensitive_tools",
            create_tool_node_with_fallback(develop_sensitive_tools),
        )
        wf.add_node(
            "develop_safe_tools",
            create_tool_node_with_fallback(develop_safe_tools),
        )

        wf.add_edge("develop_sensitive_tools", "develop_mode")
        wf.add_edge("develop_safe_tools", "develop_mode")
        wf.add_conditional_edges(
            "develop_mode",
            route_develop_mode,
            [
                "develop_sensitive_tools",
                "develop_safe_tools",
                "leave_skill",
                # END,
            ],
        )



import uuid
from datetime import datetime
from typing import Literal, Optional

from pydantic import BaseModel
from sqlmodel import JSON, Column, Field

from mtmai.models.agent import CopilotScreen
from mtmai.models.base_model import MtmBaseSqlModel
from mtmai.models.search_index import SearchRequest


class ChatThread(MtmBaseSqlModel, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    name: str = Field(nullable=True)
    userId: uuid.UUID | None = Field(
        foreign_key="user.id", index=True, nullable=False, ondelete="CASCADE"
    )
    userIdentifier: str = Field(nullable=True)
    tags: list[str] = Field(default=[], sa_column=Column(JSON))
    meta: dict | None = Field(default={}, sa_column=Column(JSON))

    # _upsert_index_elements = {"id"}


class ChatStep(MtmBaseSqlModel, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    name: str = Field(max_length=255)
    type: str = Field(max_length=255)
    thread_id: uuid.UUID = Field(...)
    parent_id: uuid.UUID | None = Field(default=None)
    disable_feedback: bool = Field(...)
    streaming: bool = Field(...)
    wait_for_answer: bool | None = Field(default=None)
    is_error: bool | None = Field(default=None)
    meta: dict | None = Field(default=None, sa_column=Column(JSON))
    tags: list[str] | None = Field(default=[], sa_column=Column(JSON))
    input: str | None = Field(default=None)
    output: str | None = Field(default=None)
    start: datetime | None = Field(default=None)
    end: datetime | None = Field(default=None)
    generation: dict | None = Field(default=None, sa_column=Column(JSON))
    show_input: str | None = Field(default=None)
    language: str | None = Field(default=None)
    indent: int | None = Field(default=None)

    _upsert_index_elements = {"id"}


class ChatElement(MtmBaseSqlModel, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    threadId: uuid.UUID | None = Field(default=None, index=True)
    type: str | None = Field(default=None)
    url: str | None = Field(default=None)
    chainlitKey: str | None = Field(default=None)
    name: str = Field(...)
    display: str | None = Field(default=None)
    objectKey: str | None = Field(default=None)
    size: str | None = Field(default=None)
    page: int | None = Field(default=None)
    language: str | None = Field(default=None)
    forId: uuid.UUID | None = Field(default=None)
    mime: str | None = Field(default=None)


class ChatElementResponse(ChatElement):
    pass


class ChatFeedback(MtmBaseSqlModel, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    for_id: uuid.UUID = Field(...)
    thread_id: uuid.UUID = Field(...)
    value: int = Field(...)
    comment: str | None = Field(default=None)

    # _upsert_index_elements = {"id"}


class Starter(BaseModel):
    """Specification for a starter that can be chosen by the user at the thread start."""

    label: str
    message: str
    icon: Optional[str] = None


class ChatProfile(MtmBaseSqlModel, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    name: str = Field(...)
    description: str = Field(...)
    icon: str | None = Field(default=None)
    default: bool | None = Field(default=False)
    starters: list[dict] | None = Field(default=None, sa_column=Column(JSON))


class ChatProfilesResponse(BaseModel):
    count: int
    data: list[ChatProfile]


class ThreadUIState(BaseModel):
    """ThreadView 的UI 状态"""

    enableChat: bool | None = False
    enableScrollToBottom: bool = True
    title: str | None = None
    description: str | None = None
    icons: str | None = None
    layout: str | None = None
    openWorkbench: bool | None = None
    openChat: bool | None = None
    theme: str | None = None
    isOpen: bool | None = None
    screens: list[CopilotScreen] = Field(default_factory=list)
    activateViewName: str | None = "/"

    # 对话输入框的位置
    inputPosition: Literal["inline", "bottom"] | None = "bottom"
    fabEnabled: bool = True
    fabIcon: str | None = None
    fabAction: str | None = None
    fabDisplayText: str | None = None
    fabDisplayIcon: str | None = None
    fabDisplayColor: str | None = None
    fabDisplayAction: str | None = None
    fabDisplayPosition: Literal["top", "bottom"] | None = "bottom"

    # play data

    playDataType: Literal["post", "demoArticle"] | None = None
    playData: dict | None = Field(default=None, sa_column=Column(JSON))


class AssisantWorkBrench(BaseModel):
    id: str
    label: str
    description: str
    icon: str | None = None


class AssisantStart(BaseModel):
    title: str
    message: str | None = None
    icon: str | None = None


class AssisantWelcome(BaseModel):
    title: str
    description: str | None = None


class AssisantWorkbenchConfig(BaseModel):
    workbenchs: list[AssisantWorkBrench]
    workbench_default: str | None = None


class AssisantContext(BaseModel):
    siteId: str | None = None
    userId: str | None = None
    threadId: str | None = None
    params: dict | None = None


class AssisantMenus(BaseModel):
    """
    应用主菜单
    """

    id: str
    label: str
    icon: str | None = None
    viewName: str | None = None
    viewProps: dict | None = None
    target: Literal["_blank", "_self", "workbench", "asider", "cmdk"] | None = None
    children: list["AssisantMenus"] | None = None


class SiderViewConfig(BaseModel):
    """
    侧边栏视图配置
    """

    id: str
    label: str
    icon: str | None = None
    viewName: str | None = None
    viewProps: dict | None = None


class ListViewItemAdditionActions(BaseModel):
    label: str
    icon: str | None = None
    action: str | None = None


class ListViewProps(BaseModel):
    isDev: bool | None = None
    # 限制只能查询单个数据类型
    dataType: str | None = None
    # q: str | None = None
    additionActions: list[ListViewItemAdditionActions] | None = None
    searchParams: SearchRequest | None = None
    variants: Literal["aside", "list", "grid"] | None = None
    enableSearch: bool | None = None
    enableFilter: bool | None = None
    enableSort: bool | None = None
    enablePagination: bool | None = None
    enableExport: bool | None = None
    enableImport: bool | None = None
    enableCreate: bool | None = None
    enableEdit: bool | None = None
    enableDelete: bool | None = None
    enableView: bool | None = None

    createFormName: str | None = None
    editFormName: str | None = None


class AssisantConfig(BaseModel):
    chatProfile: str
    logo: str | None = None  # 站点logo 暂时没用上
    welcome: AssisantWelcome | None = None
    starts: list[AssisantStart]
    workbench: AssisantWorkbenchConfig | None = None
    context: AssisantContext | None = None
    menus: list[AssisantMenus] | None = None
    siderView: SiderViewConfig | None = None



"""博客系统的 curd 操作"""

import random
import string
import uuid

from fastapi import HTTPException
from pydantic import BaseModel
from slugify import slugify
from sqlmodel import Session, delete, func, select
from sqlmodel.ext.asyncio.session import AsyncSession

from mtmai.llm.embedding import embedding_hf
from mtmai.models.blog import (
    BlogPostCreateReq,
    BlogPostUpdateRequest,
    Post,
    PostContent,
    Tag,
    TaggedItem,
)
from mtmai.models.doc import DocumentIndex
from mtmai.models.models import Document
from mtmai.models.search_index import SearchIndex
from mtmai.mtlibs.html.htmlUtils import extract_title_from_html


async def create_blog_post(
    *,
    session: AsyncSession,
    blog_post_create: BlogPostCreateReq,
    user_id: uuid.UUID | str,
) -> Post:
    """
    TODO: 改进1：
        embedding 应该用独立的方式进行，因为调用 embedding_hf 可能失败

    """
    if isinstance(user_id, str):
        user_id = uuid.UUID(user_id)
    input_data = BlogPostCreateReq.model_validate(blog_post_create)
    site_id = input_data.siteId
    if not site_id:
        raise HTTPException(status_code=400, detail="siteId is required")

    if not input_data.content:
        # 前端可以不输入任何直接创建文章
        input_data.content = "<h1>new post</h1><p>post content</p>"
    if not input_data.title:
        input_data.title = extract_title_from_html(input_data.content)
    if not input_data.title:
        input_data.title = "untitled"

    # 处理 slug
    base_slug = slugify(input_data.title)
    slug = base_slug
    while True:
        existing_post = await session.exec(select(Post).where(Post.slug == slug))
        if existing_post.first() is None:
            break
        random_string = "".join(
            random.choices(string.ascii_lowercase + string.digits, k=5)
        )
        slug = f"{base_slug}-{random_string}"
    # Create Post
    new_blog_post = Post(
        title=input_data.title,
        slug=slug,
        site_id=site_id,
        user_id=user_id,
    )

    session.add(new_blog_post)

    # 处理 tags
    if input_data.tags:
        for tag_name in input_data.tags:
            # Check if the tag already exists
            result = await session.exec(
                select(Tag).where(Tag.name == tag_name).where(Tag.site_id == site_id)
            )
            existing_tag = result.first()
            if not existing_tag:
                # If the tag doesn't exist, create a new one
                existing_tag = Tag(name=tag_name, site_id=site_id)
                session.add(existing_tag)

            # Create a TaggedItem to link the post and the tag
            tagged_item = TaggedItem(
                tag_id=existing_tag.id, item_id=new_blog_post.id, item_type="post"
            )
            session.add(tagged_item)

    post_content = PostContent(post_id=new_blog_post.id, content=input_data.content)
    session.add(post_content)

    search_index = SearchIndex(
        content_type="post",
        content_id=new_blog_post.id,
        title=new_blog_post.title,
        meta={
            # "author_id": str(new_blog_post.author_id),
            # "tags": [tag.name for tag in new_blog_post.tags],
        },
        # search_vector=generate_search_vector(post.title, post.content),
        # embedding=generate_embedding(post.title, post.content)
    )
    session.add(search_index)

    await session.commit()
    await session.refresh(new_blog_post)
    return new_blog_post


async def update_post(
    *, session: AsyncSession, post_update: BlogPostUpdateRequest
) -> Post:
    a = await session.exec(select(Post).where(Post.id == post_update.id))
    blog_post = a.one()
    if not blog_post:
        raise HTTPException(status_code=404)
    siteId = (
        blog_post.site_id
    )  # 存储siteId的值，因为后续的 session 操作会改变 blog_post

    title = post_update.title
    blog_post.title = title
    # Update slug if changed
    if post_update.slug and post_update.slug != blog_post.slug:
        existing_post = await session.exec(
            select(Post).where(Post.slug == post_update.slug)
        )
        existing_post = existing_post.first()
        if existing_post:
            raise HTTPException(status_code=400, detail="Slug already exists")
        blog_post.slug = post_update.slug

    await save_post_content(
        session=session, post_id=post_update.id, content=post_update.content
    )

    post_tags = await get_tags_by_post(session, post_update.id)
    existing_tag_names = set(tag.name for tag in post_tags)
    new_tag_names = set(post_update.tags)

    # 删除多余的tags
    await delete_tags_by_post(
        session, post_update.id, existing_tag_names - new_tag_names
    )
    await add_tags_to_posts(
        session, siteId, post_update.id, (new_tag_names - existing_tag_names)
    )

    await session.commit()
    return post_update.id


async def get_post_content(*, session: AsyncSession, post_id: uuid.UUID):
    stmt = select(PostContent).where(PostContent.post_id == post_id)
    post_content = await session.exec(stmt)
    post_content = post_content.one()
    return post_content


async def save_post_content(*, session: AsyncSession, post_id: uuid.UUID, content: str):
    blog_post = await session.exec(
        select(PostContent).where(PostContent.post_id == post_id)
    )
    blog_post = blog_post.one()
    if not blog_post:
        blog_post = PostContent(post_id=post_id, content=content)
        session.add(blog_post)
    else:
        blog_post.content = content

    await session.commit()
    await session.refresh(blog_post)
    return blog_post


async def get_tags_by_post(session: AsyncSession, post_id: uuid.UUID):
    stmt = (
        select(Tag)
        .join(TaggedItem, Tag.id == TaggedItem.tag_id)
        .where(TaggedItem.item_id == post_id)
    )
    result = await session.exec(stmt)
    return result.all()


async def delete_tags_by_post(
    session: AsyncSession, post_id: uuid.UUID, tags_to_remove: set[str]
):
    # First, get the tag IDs for the tags we want to remove
    tag_ids_to_remove = await session.exec(
        select(Tag.id)
        .where(Tag.name.in_(tags_to_remove))
        .where(
            Tag.id.in_(
                select(TaggedItem.tag_id)
                .where(TaggedItem.item_id == post_id)
                .where(TaggedItem.item_type == "post")
            )
        )
    )
    tag_ids_to_remove = tag_ids_to_remove.all()

    # Delete the TaggedItems for this post and the specified tags
    stmt = delete(TaggedItem).where(
        (TaggedItem.item_id == post_id)
        & (TaggedItem.item_type == "post")
        & (TaggedItem.tag_id.in_(tag_ids_to_remove))
    )
    await session.exec(stmt)

    # Commit the changes
    # await session.commit()


async def add_tags_to_posts(
    session: AsyncSession, site_id: uuid.UUID, post_id: uuid.UUID, tags_to_add: set[str]
):
    # First, get or create the tags
    for tag_name in tags_to_add:
        # Check if the tag already exists
        existing_tag = await session.exec(select(Tag).where(Tag.name == tag_name))
        existing_tag = existing_tag.first()

        if existing_tag:
            tag = existing_tag
        else:
            # Create a new tag if it doesn't exist
            tag = Tag(name=tag_name, site_id=site_id)
            session.add(tag)
            await session.flush()  # Flush to get the tag ID

        # Check if the post is already tagged with this tag
        existing_tagged_item = await session.exec(
            select(TaggedItem).where(
                (TaggedItem.item_id == post_id)
                & (TaggedItem.tag_id == tag.id)
                & (TaggedItem.item_type == "post")
            )
        )
        existing_tagged_item = existing_tagged_item.first()

        if not existing_tagged_item:
            # If not, create a new TaggedItem
            tagged_item = TaggedItem(item_id=post_id, tag_id=tag.id, item_type="post")
            session.add(tagged_item)

    # Commit all changes
    # await session.commit()


class DocumentQueryReq(BaseModel):
    query: str
    limit: int = 10
    offset: int = 0


async def document_query(session: Session, req: DocumentQueryReq) -> list[Document]:
    embedding_result = await embedding_hf(inputs=[req.query])
    query = (
        select(Document)
        .join(DocumentIndex, Document.id == DocumentIndex.document_id)
        .order_by(DocumentIndex.embedding.l2_distance(embedding_result[0]))
        .offset(req.offset)
        .limit(req.limit)
    )
    result = session.exec(query).all()
    return result


# async def get_post_detail(*, session: AsyncSession,post_id:str):
#     blog_post = await session.exec(select(Post).where(Post.id == post_id)).one_or_none()
#     if not blog_post:
#         raise HTTPException(status_code=404, detail="Post not found")

#     blog_post_content = await .exec(
#         select(Document).where(Document.id == blog_post.doc_id)
#     ).one_or_none()
#     # if not blog_post_content:
#     #     raise HTTPException(status_code=404, detail="Post content not found")

#     return BlogPostDetailResponse(
#         id=blog_post.id,
#         title=blog_post.title,
#         content=blog_post_content.content,
#         tags=[],  # Assuming tags are not implemented yet
#         author_id=None,  # Assuming author details are not implemented yet
#         author_avatar=None,
#     )


async def get_related_posts(db: AsyncSession, current_post: Post, limit: int = 5):
    """
    获取相关文章(未完成)
    """
    # 获取当前文章的标签
    current_tags = [tag.id for tag in current_post.tags]

    related_posts = (
        await db.query(Post)
        .join(TaggedItem)
        .filter(Post.id != current_post.id)  # 排除当前文章
        .filter(TaggedItem.tag_id.in_(current_tags))  # 匹配标签
        .group_by(Post.id)
        .order_by(func.count(TaggedItem.tag_id).desc())  # 按共享标签数量排序
        .limit(limit)
        .all()
    )
    return related_posts



from dataclasses import dataclass
from typing import Any

from autogen_core import DefaultInterventionHandler, MessageContext


@dataclass
class Message:
    content: Any


@dataclass
class Termination:
    reason: str


class TerminationHandler(DefaultInterventionHandler):
    def __init__(self) -> None:
        self._termination_value: Termination | None = None

    async def on_publish(self, message: Any, *, message_context: MessageContext) -> Any:
        if isinstance(message, Termination):
            self._termination_value = message
        return message

    @property
    def termination_value(self) -> Termination | None:
        return self._termination_value

    @property
    def has_terminated(self) -> bool:
        return self._termination_value is not None



"""博客系统的 curd 操作"""

from datetime import datetime

from fastapi import HTTPException
from pydantic import BaseModel
from sqlmodel import select

from mtmai.db.db import get_async_session
from mtmai.models.logitems import LogItem, LogItemListResponse


class LogItemCreateReq(BaseModel):
    app: str
    text: str
    level: int = 3
    resource_id: str | None = None


async def create_log_item(
    log_item_create: LogItemCreateReq,
) -> LogItem:
    input_data = LogItemCreateReq.model_validate(log_item_create)

    if not input_data.text:
        raise HTTPException(status_code=400, detail="text is required")
    if not input_data.app:
        input_data = "sys"

    async with get_async_session() as session:
        new_log_item = LogItem(**input_data.model_dump(), title="text")
        session.add(new_log_item)
        await session.commit()
        await session.refresh(new_log_item)
    return new_log_item


async def create_context_logger():
    pass


async def list_log_items(app: str, resource_id: str | None = None):
    async with get_async_session() as session:
        query = select(LogItem).where(
            LogItem.app == app,
            LogItem.is_deleted == False,  # noqa: E712
            LogItem.expire_at > datetime.now(),
            LogItem.resource_id == resource_id,
        )
        if resource_id:
            query = query.where(LogItem.resource_id == resource_id)
        result = await session.exec(query)
        items = result.all()

        return LogItemListResponse(items=items, total=len(items))



from __future__ import annotations

from datetime import datetime
from typing import Any, ClassVar, Generic, TypeVar

import sqlalchemy as sa
from pydantic import BaseModel, ConfigDict
from sqlalchemy.dialects.postgresql import Insert, insert
from sqlmodel import Field, SQLModel


class MtmBaseSqlModel(SQLModel):
    """Base class for SQL models.
    用途
         1: 作为 基于 SQLModel 的自定义基类
         2: 添加 upsert的支持, 参考: https://github.com/dan1elt0m/sadel/blob/main/sadel/base.py
    """

    model_config = ConfigDict(validate_assignment=True)  # pyright: ignore
    created_at: datetime | None = Field(
        default=None,
        sa_type=sa.DateTime(timezone=True),  # type: ignore
        sa_column_kwargs={"server_default": sa.func.now()},
        description=(
            "The date and time the record was created. "
            "Field is optional and not needed when instantiating a new record. "
            "It will be automatically set when the record is created in the database."
        ),
    )

    updated_at: datetime | None = Field(
        default=None,
        sa_type=sa.DateTime(timezone=True),  # pyright: ignore
        sa_column_kwargs={"onupdate": sa.func.now(), "server_default": sa.func.now()},
        description=(
            "The date and time the record was updated. "
            "Field is optional and not needed when instantiating a new record. "
            "It will be automatically set when the record is created in the database."
        ),
    )

    # Specifies the set of index elements which represent the ON CONFLICT target
    _upsert_index_elements: ClassVar[set[str]] = set()

    # Specifies the set of fields to exclude from updating in the resulting
    # UPSERT statement
    _upsert_exclude_fields: ClassVar[set[str]] = set()

    # Common fields which we should exclude when updating.
    _default_upsert_exclude_fields: ClassVar[set[str]] = {"created_at"}

    @classmethod
    async def upsert(cls, item: MtmBaseSqlModel, session: sa.orm.Session):
        """upserts a single item"""
        stmt = cls._get_upsert_statement(item)
        await session.exec(stmt)
        await session.commit()

    @classmethod
    async def batch_upsert(cls, items: list[MtmBaseSqlModel], session: sa.orm.Session):
        """Batch upserts a list of items."""
        for item in items:
            stmt = cls._get_upsert_statement(item)
            await session.exec(stmt)
        await session.commit()

    @classmethod
    def _get_upsert_statement(cls, item: MtmBaseSqlModel) -> Insert:
        """Returns an UPSERT statement for a single item."""
        if not cls._upsert_index_elements:
            raise ValueError("No upsert index elements specified for the model.")

        to_insert = item.model_dump()
        to_insert["created_at"] = (
            sa.func.now()
        )  # set manually, because on_conflict_do_update doesn't trigger default oninsert
        to_update = cls._get_record_to_update(to_insert)
        stmt = insert(cls).values(to_insert)
        return stmt.on_conflict_do_update(
            index_elements=cls._upsert_index_elements,
            set_=to_update,
        )

    @classmethod
    def _get_record_to_update(cls, record: dict[str, Any]) -> dict[str, Any]:
        """Returns a record to be upserted taking into account the excluded fields."""
        exclude_fields = cls._upsert_exclude_fields.copy()
        exclude_fields.update(cls._default_upsert_exclude_fields)

        to_update = record.copy()

        for field in exclude_fields:
            _ = to_update.pop(field, None)  # pyright: ignore

        to_update["updated_at"] = (
            sa.func.now()
        )  # set manually. on_conflict_do_update doesn't trigger onupdate

        return to_update  # type: ignore


class CommonResultRequest(BaseModel):
    q: str | None = Field(default=None, max_length=255)
    skip: int = Field(default=0)
    limit: int = Field(default=10)


class CommonResultResponse(BaseModel):
    error: str = Field(default="")
    data: dict = Field(default={})
    message: str = Field(default="")


T = TypeVar("T")


class ListResponse(BaseModel, Generic[T]):
    count: int
    items: list[T]



# import logging
# import uuid
# from typing import Any, Dict, List, Union

from contextlib import asynccontextmanager

from psycopg_pool import AsyncConnectionPool

# from sqlalchemy import text
# from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from sqlmodel import Session, create_engine, select
from sqlmodel.ext.asyncio.session import AsyncSession

from mtmai.core.config import settings
from mtmai.crud import curd
from mtmai.models.models import User, UserCreate


def init_db(session: Session) -> None:
    user = session.exec(
        select(User).where(User.email == settings.FIRST_SUPERUSER)
    ).first()
    if not user:
        user_in = UserCreate(
            email=settings.FIRST_SUPERUSER,
            password=settings.FIRST_SUPERUSER_PASSWORD,
            is_superuser=True,
        )
        user = curd.create_user(session=session, user_create=user_in)


engine = None


def fix_conn_str(conn_str: str) -> str:
    if not str(conn_str).startswith("postgresql+psycopg"):
        conn_str = str(conn_str).replace("postgresql", "postgresql+psycopg")
    return conn_str


def get_engine():
    global engine
    if engine is not None:
        return engine
    if settings.MTMAI_DATABASE_URL is None:
        raise ValueError("MTMAI_DATABASE_URL environment variable is not set")  # noqa: EM101, TRY003
    return create_engine(
        settings.MTMAI_DATABASE_URL,
        connect_args={"sslmode": "require"},
        pool_recycle=300,
    )


# engine = get_engine()


# 全局连接池对象
pool: AsyncConnectionPool | None = None


async def get_checkpointer():
    from langgraph.checkpoint.postgres.aio import AsyncPostgresSaver

    global pool
    if not pool or pool.closed:
        connection_kwargs = {
            "autocommit": True,
            "prepare_threshold": 0,
        }
        pool = AsyncConnectionPool(
            conninfo=settings.MTMAI_DATABASE_URL,
            max_size=20,
            kwargs=connection_kwargs,
        )
        await pool.open()
    checkpointer = AsyncPostgresSaver(pool)
    yield checkpointer


async_engine: AsyncEngine | None = None


def get_async_engine():
    global async_engine
    if async_engine is not None:
        return async_engine
    if settings.MTMAI_DATABASE_URL is None:
        raise ValueError("DATABASE_URL environment variable is not set")  # noqa: EM101, TRY003

    return create_async_engine(
        fix_conn_str(settings.MTMAI_DATABASE_URL),
        #    echo=True,# echo 会打印所有sql语句，影响性能
        future=True,
    )


@asynccontextmanager
async def get_async_session():
    engine = get_async_engine()
    async with AsyncSession(engine) as session:
        try:
            yield session
        finally:
            await session.close()



# coding: utf-8

"""
    Mtmai API

    The Mtmai API

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from mtmai.gomtmclients.rest.models.agent_finish import AgentFinish
from mtmai.gomtmclients.rest.models.agent_node_output import AgentNodeOutput
from mtmai.gomtmclients.rest.models.agent_state import AgentState
from mtmai.gomtmclients.rest.models.agent_step import AgentStep
from mtmai.gomtmclients.rest.models.api_resource_meta import APIResourceMeta
from typing import Optional, Set
from typing_extensions import Self

class AgentNode(BaseModel):
    """
    AgentNode
    """ # noqa: E501
    metadata: APIResourceMeta
    title: StrictStr = Field(description="agent 节点名称, 或者作为工具名称")
    description: StrictStr = Field(description="agent 节点描述, 或者作为工具描述")
    type: Optional[StrictStr] = Field(default=None, description="节点类型，决定了由哪个agent来进行调度")
    config: Optional[Dict[str, Any]] = Field(default=None, description="agent 的配置", alias="Config")
    state: Optional[AgentState] = None
    steps: List[AgentStep] = Field(description="agent 节点执行步骤, 一般表示之前执行的步骤")
    finish: Optional[AgentFinish] = None
    parent_id: Optional[StrictStr] = Field(default=None, description="上级节点", alias="parentId")
    agent_node_output: Optional[AgentNodeOutput] = Field(default=None, alias="agentNodeOutput")
    tools: Optional[StrictStr] = None
    memory_id: Optional[StrictStr] = Field(default=None, description="记忆ID，表示这个agent的记忆", alias="memoryId")
    input: Optional[StrictStr] = Field(default=None, description="输入")
    output: Optional[StrictStr] = Field(default=None, description="输出")
    __properties: ClassVar[List[str]] = ["metadata", "title", "description", "type", "Config", "state", "steps", "finish", "parentId", "agentNodeOutput", "tools", "memoryId", "input", "output"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of AgentNode from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of metadata
        if self.metadata:
            _dict['metadata'] = self.metadata.to_dict()
        # override the default output from pydantic by calling `to_dict()` of state
        if self.state:
            _dict['state'] = self.state.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in steps (list)
        _items = []
        if self.steps:
            for _item_steps in self.steps:
                if _item_steps:
                    _items.append(_item_steps.to_dict())
            _dict['steps'] = _items
        # override the default output from pydantic by calling `to_dict()` of finish
        if self.finish:
            _dict['finish'] = self.finish.to_dict()
        # override the default output from pydantic by calling `to_dict()` of agent_node_output
        if self.agent_node_output:
            _dict['agentNodeOutput'] = self.agent_node_output.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of AgentNode from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "metadata": APIResourceMeta.from_dict(obj["metadata"]) if obj.get("metadata") is not None else None,
            "title": obj.get("title"),
            "description": obj.get("description"),
            "type": obj.get("type"),
            "Config": obj.get("Config"),
            "state": AgentState.from_dict(obj["state"]) if obj.get("state") is not None else None,
            "steps": [AgentStep.from_dict(_item) for _item in obj["steps"]] if obj.get("steps") is not None else None,
            "finish": AgentFinish.from_dict(obj["finish"]) if obj.get("finish") is not None else None,
            "parentId": obj.get("parentId"),
            "agentNodeOutput": AgentNodeOutput.from_dict(obj["agentNodeOutput"]) if obj.get("agentNodeOutput") is not None else None,
            "tools": obj.get("tools"),
            "memoryId": obj.get("memoryId"),
            "input": obj.get("input"),
            "output": obj.get("output")
        })
        return _obj





from typing import Annotated

import structlog
from fastapi import APIRouter, BackgroundTasks, HTTPException, Query, Request
from fastapi.encoders import jsonable_encoder
from fastapi.responses import ORJSONResponse
from pydantic import BaseModel
from websockets.exceptions import ConnectionClosedOK

from mtmai import analytics

# from mtmai.biz import biz_scheule
from mtmai.core.event import emit_flow_event
from mtmai.crud import crud_sysitem, crud_task, curd
from mtmai.deps import AsyncSessionDep, CurrentUser
from mtmai.executor.factory import AsyncExecutorFactory
from mtmai.forge import app
from mtmai.forge.sdk.schemas.tasks import OrderBy, SortDirection, TaskStatus
from mtmai.models.base_model import CommonResultResponse
from mtmai.models.task import MtTask, MtTaskStatus

LOG = structlog.get_logger()
websocket_router = APIRouter()
router = APIRouter()


@router.get("/tasks", response_model=list[MtTask])
async def task_list(
    *,
    session: AsyncSessionDep,
    user: CurrentUser,
    page: int = Query(1, ge=1),
    page_size: int = Query(10, ge=1),
    task_status: Annotated[list[TaskStatus] | None, Query()] = None,
    workflow_run_id: Annotated[str | None, Query()] = None,
    # current_org: Organization = Depends(org_auth_service.get_current_org),
    only_standalone_tasks: bool = Query(False),
    sort: OrderBy = Query(OrderBy.created_at),
    order: SortDirection = Query(SortDirection.desc),
):
    analytics.capture("skyvern-oss-agent-mttask-list")
    # 1: TODO: 检测执行权限
    # await PermissionCheckerFactory.get_instance().check(current_org)

    tasks = await crud_task.get_mttasks(
        page=page,
        page_size=page_size,
        # task_status=task_status,
        # workflow_run_id=workflow_run_id,
    )

    return ORJSONResponse([jsonable_encoder(task) for task in tasks])


class CreateTaskRequest(BaseModel):
    task_type: str
    params: dict


class CreateTaskResponse(BaseModel):
    task_id: str


@router.post("/task", response_model=CreateTaskResponse)
async def create_mttask(
    *,
    session: AsyncSessionDep,
    user: CurrentUser,
    req: CreateTaskRequest,
    request: Request,
    background_tasks: BackgroundTasks,
):
    """
    创建任务，并执行。
    """
    analytics.capture(
        "skyvern-oss-agent-mttask-create", data={"task_type": req.task_type}
    )

    # 1: TODO: 检测执行权限
    # await PermissionCheckerFactory.get_instance().check(current_org)

    # 数据库创建任务
    new_task = await crud_task.create_task(
        session=session, task_create=req, owner_id=user.id
    )

    # 执行任务
    await AsyncExecutorFactory.get_executor().execute_graph(
        request=request,
        background_tasks=background_tasks,
        graph_id="taskrunner",  # todo
        thread_id=new_task.id,  # todo
        # task_id=req.task_type,
        # organization_id=current_org.organization_id,
        # max_steps_override=x_max_steps_override,
        # api_key=get_api_key,
    )
    return CreateTaskResponse(task_id=new_task.id)


@router.get("/{mttask_id}", response_model=MtTask)
async def get_mttask(
    *,
    session: AsyncSessionDep,
    current_user: CurrentUser,
    mttask_id: str,
):
    """
    Get task by id.
    """
    item = await crud_task.mttask_get_by_id(session=session, mttask_id=mttask_id)
    if not item or item.owner_id != current_user.id:
        raise HTTPException(status_code=404, detail="not found")
    return item


class DrowndownSelectItem(BaseModel):
    id: str
    value: str
    title: str
    description: str


class SelectItemsResponse(BaseModel):
    items: list[DrowndownSelectItem]
    count: int


@router.post("/task_types", response_model=SelectItemsResponse)
async def task_types(
    *,
    session: AsyncSessionDep,
    current_user: CurrentUser,
):
    """
    Get task types.
    """

    sysitems = await crud_sysitem.get_sys_items(session=session, type="task_type")
    if sysitems:
        items = []
        for sysitem in sysitems:
            items.append(
                DrowndownSelectItem(
                    id=sysitem.key,
                    title=sysitem.description,
                    description=sysitem.description,
                    value=sysitem.value,
                )
            )
        return SelectItemsResponse(items=items, count=len(items))

    raise HTTPException(status_code=404, detail="Task types not found")


class MttaskUpdateStatusRequest(BaseModel):
    mttask_id: str
    status: str


@router.post("/mttask_update_status", response_model=CommonResultResponse)
async def mttask_update_status(
    *,
    session: AsyncSessionDep,
    current_user: CurrentUser,
    item_in: MttaskUpdateStatusRequest,
):
    """
    更新任务状态
    """
    mttask = await crud_task.mttask_get_by_id(
        session=session, mttask_id=item_in.mttask_id
    )
    if not mttask:
        raise HTTPException(status_code=404, detail="not found")
    if mttask.owner_id != current_user.id:
        raise HTTPException(status_code=403, detail="forbidden")

    # 状态对比
    if mttask.status == item_in.status:
        return CommonResultResponse(
            data={"status": "success"}, message="更新任务状态成功"
        )

    # 如果状态不一样
    if mttask.status == MtTaskStatus.NEW:
        if item_in.status == "pending":
            event_result = await emit_flow_event(
                event="mtmai.mttask.update_status",
                resource_id=str(mttask.id),
                # data={"mttask_id": str(mttask.id)},
            )
            logger.info(event_result)
    if mttask.status == "pending":
        if item_in.status == "running":
            await emit_flow_event(
                event="mtmai.mttask.update_status",
                resource_id=str(mttask.id),
                # data={"mttask_id": str(mttask.id)},
            )
            session.add(mttask)
            await session.commit()
            return CommonResultResponse(
                data={"status": "success"}, message="更新任务状态成功"
            )

        elif item_in.status == "success":
            pass
        elif item_in.status == "failed":
            pass
        # else:
        #     raise HTTPException(status_code=400, detail="invalid status")
    # await crud_task.mttask_update_state(
    #     session=session, mttask_id=item_in.mttask_id, status=item_in.status
    # )
    raise HTTPException(status_code=400, detail="invalid status")


# @router.get("/schedule/{id}", response_model=ScheduleDetailPublic)
# async def schedule_get(
#     *,
#     session: AsyncSessionDep,
#     current_user: CurrentUser,
#     id: str,
# ) -> Any:
#     try:
#         return await crud_task.get_schedule(session, id, current_user.id)
#     except Exception as e:
#         logger.error(e)
#         raise HTTPException(status_code=404, detail="not found")


# @router.post("/schedule/list", response_model=ListResponse[SiteAutoItemPublic])
# async def schedule_list(
#     session: AsyncSessionDep,
#     current_user: OptionalUserDep,
#     req: ScheduleListRequest,
# ):
#     """
#     Retrieve site auto items with pagination.
#     """
#     items, count = await crud_task.list_schedule(session, req, current_user.id)

#     return ListResponse(items=jsonable_encoder(items), count=count)


# @router.post("/schedule/create", response_model=ScheduleCreateResponse)
# async def schedule_create(
#     *,
#     session: AsyncSessionDep,
#     current_user: CurrentUser,
#     req: ScheduleCreateRequest,
# ) -> Any:
#     return await crud_task.create_schedule(session, req, user_id=current_user.id)


# @router.put("/schedule/update/{id}", response_model=ScheduleUpdateResponse)
# async def schedule_update(
#     *,
#     session: AsyncSessionDep,
#     current_user: CurrentUser,
#     id: str,
#     item_in: ScheduleUpdateRequest,
# ) -> Any:
#     item_in.id = uuid.UUID(id)
#     update_result = await crud_task.update_schedule(session, item_in, current_user.id)
#     return update_result


@router.delete("/chat_profile/{id}", response_model=CommonResultResponse)
async def cht_profile_delete(
    *,
    session: AsyncSessionDep,
    current_user: CurrentUser,
    id: str,
):
    """根据调度配置，立即生成一个运行任务"""
    new_task = await biz_scheule.delete_chat_profile(id)
    return new_task


class TaskArtifact(BaseModel):
    artifact_type: str
    artifact_data: str


class TaskStatusResponse(BaseModel):
    task_id: str
    status: TaskStatus
    artifact: TaskArtifact | None = None


class TaskStatusRequest(BaseModel):
    task_id: str


@router.post("/task_status_v2", response_model=TaskStatusResponse)
async def status_v2(
    req: TaskStatusRequest,
    current_user: CurrentUser,
    apikey: str | None = None,
    token: str | None = None,
):
    """
    Get the status of a task, 专用于 TaskActions 界面中的状态沦陷（放弃websocket形式）
    """
    try:
        # organization = await get_current_org(x_api_key=apikey, authorization=token)
        org = await curd.get_organization_by_user_id(current_user.id)
        organization_id = org.organization_id
    except Exception:
        # LOG.exception("Error while getting organization", task_id=task_a)
        try:
            raise HTTPException(status_code=401, detail="Invalid credential provided")
        except ConnectionClosedOK:
            LOG.info(
                "ConnectionClosedOK error while sending invalid credential message"
            )
        return

    task = await app.DATABASE.get_task(
        task_id=req.task_id, organization_id=organization_id
    )
    return TaskStatusResponse(
        task_id=task.task_id,
        status=task.status,
        # artifact=task.artifact,
    )


class CreateArticleTaskParamsRequest(BaseModel):
    prompt: str


class ArticleGenTaskInputs(BaseModel):
    prompt: str
    topic: str
    gen_strategy: str
    gen_count: int


@router.post("/create_article_task_params", response_model=ArticleGenTaskInputs)
async def create_article_task_params(
    req: CreateArticleTaskParamsRequest,
    current_user: CurrentUser,
):
    """
    创建文章任务参数
    """
    return ArticleGenTaskInputs(
        prompt=req.prompt,
        topic=req.prompt,
        gen_strategy="",
        gen_count=1,
    )



from mtmlib import mtutils
from mtmlib.env import in_jupyter_notebook
from mtmlib.mtutils import bash


def init_env():
    """
    根据环境进行必要的初始化
    """
    if in_jupyter_notebook():
        bash("pip install ipykernel")
        mtutils.install_packages_if_missing(["ipykernel"])
        # 提示，在colab 可以创建自定义的虚拟环境， 在 Colab 的 "Runtime" > "Change runtime type" 中选择 mtmtrain_env 作为内核。
        # bash("python -m ipykernel install --user --name=mtmtrain_env")
        bash("pip install -U --no-cache-dir mtmlib")



from functools import cache
from typing import Any

from opentelemetry import trace
from opentelemetry.context import Context
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace import NoOpTracerProvider, Tracer
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

from mtmaisdk.loader import ClientConfig

OTEL_CARRIER_KEY = "__otel_carrier"


@cache
def create_tracer(config: ClientConfig) -> Tracer:
    ## TODO: Figure out how to specify protocol here
    resource = Resource(
        attributes={SERVICE_NAME: config.otel_service_name or "hatchet.run"}
    )

    if config.otel_exporter_oltp_endpoint and config.otel_exporter_oltp_headers:
        processor = BatchSpanProcessor(
            OTLPSpanExporter(
                endpoint=config.otel_exporter_oltp_endpoint,
                headers=config.otel_exporter_oltp_headers,
            ),
        )

        ## If tracer provider is already set, we don't need to override it
        if not isinstance(trace.get_tracer_provider(), TracerProvider):
            trace_provider = TracerProvider(resource=resource)
            trace_provider.add_span_processor(processor)
            trace.set_tracer_provider(trace_provider)
    else:
        if not isinstance(trace.get_tracer_provider(), NoOpTracerProvider):
            trace.set_tracer_provider(NoOpTracerProvider())

    return trace.get_tracer(__name__)


def create_carrier() -> dict[str, str]:
    carrier: dict[str, str] = {}
    TraceContextTextMapPropagator().inject(carrier)

    return carrier


def inject_carrier_into_metadata(
    metadata: dict[Any, Any], carrier: dict[str, str]
) -> dict[Any, Any]:
    if carrier:
        metadata[OTEL_CARRIER_KEY] = carrier

    return metadata


def parse_carrier_from_metadata(metadata: dict[str, Any] | None) -> Context | None:
    if not metadata:
        return None

    return (
        TraceContextTextMapPropagator().extract(_ctx)
        if (_ctx := metadata.get(OTEL_CARRIER_KEY))
        else None
    )



agentTopics = [
  'Business',
  'Marketing',
  'Finance',
  'Startups',
  'Networking',
  'Leadership',
  'Strategy',
  'Branding',
  'Analytics',
  'Growth',
  'Drawing',
  'Painting',
  'Design',
  'Photography',
  'Writing',
  'Sculpting',
  'Animation',
  'Sketching',
  'Crafting',
  'Calligraphy',
  'Mindset',
  'Productivity',
  'Motivation',
  'Education',
  'Learning',
  'Skills',
  'Success',
  'Wellness',
  'Goals',
  'Inspiration',
  'Fashion',
  'Travel',
  'Food',
  'Fitness',
  'Health',
  'Beauty',
  'Home',
  'Decor',
  'Hobbies',
  'Parenting',
  'Tech',
  'Gadgets',
  'AI',
  'Coding',
  'Software',
  'Innovation',
  'Apps',
  'Gaming',
  'Robotics',
  'Security',
  'Music',
  'Movies',
  'Sports',
  'Books',
  'Theater',
  'Comedy',
  'Dance',
  'Celebrities',
  'Culture',
  'Gaming',
  'Environment',
  'Equality',
  'Activism',
  'Justice',
  'Diversity',
  'Sustainability',
  'Inclusion',
  'Awareness',
  'Charity',
  'Peace',
  'Holidays',
  'Festivities',
  'Seasons',
  'Trends',
  'Celebrations',
  'Anniversaries',
  'Milestones',
  'Memories',
  'Promotions',
  'Updates',
];


# flake8: noqa

# import apis into api package
from mtmai.gomtmclients.rest.api.api_token_api import APITokenApi
from mtmai.gomtmclients.rest.api.event_api import EventApi
from mtmai.gomtmclients.rest.api.github_api import GithubApi
from mtmai.gomtmclients.rest.api.healthcheck_api import HealthcheckApi
from mtmai.gomtmclients.rest.api.log_api import LogApi
from mtmai.gomtmclients.rest.api.metadata_api import MetadataApi
from mtmai.gomtmclients.rest.api.rate_limits_api import RateLimitsApi
from mtmai.gomtmclients.rest.api.sns_api import SNSApi
from mtmai.gomtmclients.rest.api.slack_api import SlackApi
from mtmai.gomtmclients.rest.api.step_run_api import StepRunApi
from mtmai.gomtmclients.rest.api.tenant_api import TenantApi
from mtmai.gomtmclients.rest.api.user_api import UserApi
from mtmai.gomtmclients.rest.api.worker_api import WorkerApi
from mtmai.gomtmclients.rest.api.workflow_api import WorkflowApi
from mtmai.gomtmclients.rest.api.workflow_run_api import WorkflowRunApi
from mtmai.gomtmclients.rest.api.admin_api import AdminApi
from mtmai.gomtmclients.rest.api.agent_api import AgentApi
from mtmai.gomtmclients.rest.api.artifact_api import ArtifactApi
from mtmai.gomtmclients.rest.api.assisant_api import AssisantApi
from mtmai.gomtmclients.rest.api.blog_api import BlogApi
from mtmai.gomtmclients.rest.api.browser_api import BrowserApi
from mtmai.gomtmclients.rest.api.chat_api import ChatApi
from mtmai.gomtmclients.rest.api.default_api import DefaultApi
from mtmai.gomtmclients.rest.api.endpoint_api import EndpointApi
from mtmai.gomtmclients.rest.api.env_api import EnvApi
from mtmai.gomtmclients.rest.api.frontend_api import FrontendApi
from mtmai.gomtmclients.rest.api.galleries_api import GalleriesApi
from mtmai.gomtmclients.rest.api.hf_api import HfApi
from mtmai.gomtmclients.rest.api.kv_api import KvApi
from mtmai.gomtmclients.rest.api.llm_api import LlmApi
from mtmai.gomtmclients.rest.api.mtmai_api import MtmaiApi
from mtmai.gomtmclients.rest.api.platform_api import PlatformApi
from mtmai.gomtmclients.rest.api.platform_account_api import PlatformAccountApi
from mtmai.gomtmclients.rest.api.post_api import PostApi
from mtmai.gomtmclients.rest.api.posts_api import PostsApi
from mtmai.gomtmclients.rest.api.prompt_api import PromptApi
from mtmai.gomtmclients.rest.api.proxy_api import ProxyApi
from mtmai.gomtmclients.rest.api.session_api import SessionApi
from mtmai.gomtmclients.rest.api.sessions_api import SessionsApi
from mtmai.gomtmclients.rest.api.site_api import SiteApi
from mtmai.gomtmclients.rest.api.site_host_api import SiteHostApi
from mtmai.gomtmclients.rest.api.subscribe_api import SubscribeApi
from mtmai.gomtmclients.rest.api.team_api import TeamApi
from mtmai.gomtmclients.rest.api.teams_api import TeamsApi




import uuid
from datetime import datetime

from sqlmodel import Field, SQLModel


class ProjectBase(SQLModel):
    title: str | None = Field(default=None, max_length=255)
    description: str | None = Field(default=None, max_length=255)
    created_at: datetime = Field(default=datetime.now())
    updated_at: datetime = Field(default=datetime.now())
    enabled: bool | None = Field(default=True)
    owner_id: uuid.UUID | None = Field(
        foreign_key="user.id", index=True, nullable=False, ondelete="CASCADE"
    )



import httpx
from bs4 import BeautifulSoup


async def get_readable_text(url: str) -> str:
    """获取 AI 可读文本"""
    try:
        response = httpx.get(url)
        response.raise_for_status()  # Raises a HTTPError if the status is 4xx, 5xx

        # Parse the webpage with BeautifulSoup
        soup = BeautifulSoup(response.text, "html.parser")

        # Define the tags to be removed
        tags = [
            "a",
            "img",
            "script",
            "style",
            "svg",
            "iframe",
            "canvas",
            "video",
            "audio",
            "map",
            "noscript",
        ]

        # Remove each defined tag
        for tag in tags:
            for item in soup.find_all(tag):
                item.decompose()

        # Extract text from the parsed HTML
        text = soup.get_text()

        # Remove extra whitespace
        text = " ".join(text.split())
        # completion = self.anthropic.completions.create(
        #     model="claude-2",
        #     max_tokens_to_sample=1000000,
        #     prompt=f"{HUMAN_PROMPT} Please provide a exhaustive and concise summary of the following (formatted nicely in markdown): {text} {AI_PROMPT}",
        #     stream=False,
        # )
        # return completion.completion
        return text
    except httpx.HTTPError as e:
        return f"A HTTP error occurred: {str(e)}"
    except httpx.RequestException as e:
        return f"A request exception occurred: {str(e)}"
    except Exception as e:
        return f"An error occurred: {str(e)}"

    except Exception as e:
        raise e



import unittest



# 测试数据
test_texts = [
    "Scientists discover a new planet in our solar system, challenging existing astronomical theories.",
    "The stock market experiences a significant drop in value, causing concerns among investors.",
    "Apple announces the launch of a new MacBook Pro with enhanced performance and features.",
    "Breaking news: Political unrest in the capital city as protests escalate in response to government actions.",
    "Microsoft releases a major update for Windows 10, introducing new functionalities and improvements.",
    "Researchers find a potential cure for a common disease, marking a significant breakthrough in medical science.",
    "Tesla unveils its latest electric car model, promising longer battery life and advanced autonomous driving capabilities.",
    "International summit discusses climate change solutions, bringing together leaders from around the world.",
    "Google introduces a new feature in its search engine to provide more accurate and personalized results.",
    "Economic indicators point to a recovery in the job market, with increasing employment opportunities.",
    "SpaceX successfully launches a satellite into orbit, expanding global communication capabilities.",
    "Tech company invests in renewable energy research to promote sustainable and eco-friendly solutions.",
    "New study reveals the impact of social media on mental health, raising awareness about online well-being.",
    "Facebook announces stricter policies for content moderation to address issues related to misinformation and harmful content.",
    "Global health organization issues a warning about a new virus outbreak, urging precautionary measures.",
    "Amazon expands its drone delivery program to new regions, aiming to improve the efficiency of package delivery.",
    "Government introduces regulations to address cybersecurity threats, safeguarding national digital infrastructure.",
    "Automaker recalls vehicles due to safety concerns, prioritizing consumer safety and satisfaction.",
    "Artificial intelligence breakthrough in medical diagnostics shows promising results for early disease detection.",
    "Virtual reality technology used in innovative educational programs to enhance learning experiences for students.",
]

# 科技类别的测试文本
tech_test_texts = [
    "Apple's latest iPhone features cutting-edge technology and a sleek design.",
    "SpaceX announces plans for a mission to Mars, revolutionizing space exploration.",
    "Google develops a new artificial intelligence algorithm for image recognition.",
    "Microsoft unveils a breakthrough in quantum computing, paving the way for advanced computational capabilities.",
    "Tesla introduces a self-driving electric car with advanced safety features.",
    "Researchers create a new material with potential applications in renewable energy technology.",
    "Amazon launches a state-of-the-art data center for cloud computing services.",
    "Virtual reality startup raises significant funding for the development of immersive gaming experiences.",
    "Artificial intelligence company achieves a major milestone in natural language processing.",
    "NASA collaborates with private tech companies to develop next-generation space technologies.",
    "Tech giant invests in the development of a high-speed, low-latency 5G network.",
    "Blockchain technology gains widespread adoption in the finance and healthcare sectors.",
    "Electric vehicle startup introduces a groundbreaking energy-efficient charging system.",
    "Robotics company unveils a humanoid robot with advanced capabilities for various industries.",
    "Drones equipped with AI are used for innovative applications in agriculture and environmental monitoring.",
    "Augmented reality devices become increasingly popular for educational and entertainment purposes.",
    "Biotechnology firm pioneers gene-editing techniques for potential medical breakthroughs.",
    "Cybersecurity company develops an advanced threat detection system to protect against cyber attacks.",
    "Renewable energy startup achieves a major milestone in the development of efficient solar panels.",
    "Innovative tech solutions contribute to the advancement of smart cities around the world.",
]

# 食品类别的测试文本
food_test_texts = [
    "New restaurant opens with a menu featuring diverse and delicious cuisines.",
    "Cooking show introduces a unique recipe for a popular dessert.",
    "Food festival showcases a variety of local and international dishes.",
    "Health benefits of a balanced diet are discussed in a nutrition seminar.",
    "Celebrity chef launches a new line of gourmet products.",
    "Vegan and gluten-free options available at a trendy new cafe.",
    "Street food vendors offer tasty and affordable snacks in the city center.",
    "Review of a new cookbook with innovative and easy-to-follow recipes.",
    "Traditional recipes passed down through generations shared in a culinary blog.",
    "Food enthusiasts gather for a community potluck with homemade dishes.",
    "Exploration of exotic ingredients and flavors in a travel and food documentary.",
    "Tips for sustainable and eco-friendly food choices in a lifestyle magazine.",
    "Cooking competition features skilled chefs showcasing their culinary expertise.",
    "Local farmers market promotes fresh and locally sourced produce.",
    "Artisanal chocolate brand recognized for its quality and ethical sourcing.",
    "Interactive cooking class teaches participants how to prepare a gourmet meal.",
    "Food and wine pairing event highlights the perfect combinations for a memorable dining experience.",
    "Review of a popular street food vendor known for its signature dishes.",
    "Exploration of unique street food offerings in different cities around the world.",
    "New technology in food production aims to address global challenges and enhance sustainability.",
]

class TestClasssify(unittest.TestCase):
    def test_hello_world(self):
        # result = hello_world()
        # self.assertEqual(result, "Hello, World!")
        print ("hello")




if __name__ == '__main__':
    unittest.main()



import logging

logger = logging.getLogger()


def format_doc(doc, max_length=1000):
    related = "- ".join(doc.metadata["categories"])
    return f"### {doc.metadata['title']}\n\nSummary: {doc.page_content}\n\nRelated\n{related}"[
        :max_length
    ]


def format_docs(docs):
    return "\n\n".join(format_doc(doc) for doc in docs)


class MtmTopicDocRetriever:
    """根据给定的主题列表，从网络获取相关主题的文章"""

    def __init__(self):
        pass

    async def retrive(self, topics: list[str], max_retry=3, min_docs_required=1):
        logger.info(
            f"MtmTopicDocRetriever 开始检索相关文档:{min_docs_required}, {topics}"
        )
        from langchain_community.retrievers import WikipediaRetriever

        wikipedia_retriever = WikipediaRetriever(
            load_all_available_meta=True, top_k_results=1
        )
        all_docs = []
        retry_count = 0

        while retry_count < max_retry and len(all_docs) < min_docs_required:
            retrieved_docs = await wikipedia_retriever.abatch(
                topics, return_exceptions=True
            )

            for docs in retrieved_docs:
                if isinstance(docs, BaseException):
                    logger.error(f"wikipedia 检索失败: {docs}")
                    continue
                if len(docs) == 0:
                    logger.error(f"wikipedia 检索失败,0 文档: {docs}")
                    continue
                all_docs.extend(docs)

            if len(all_docs) < min_docs_required:
                logger.warning(
                    f"重试获取文档，当前重试次数: {retry_count + 1}，已获取 {len(all_docs)} 个文档，目标数量: {min_docs_required}"
                )
                retry_count += 1

        if len(all_docs) < min_docs_required:
            logger.error(
                f"在{max_retry}次重试后仍未获取到足够的文档。目标数量: {min_docs_required}，实际获取: {len(all_docs)}"
            )
            if len(all_docs) == 0:
                return None

        formatted = format_docs(all_docs)

        return formatted



"""
为前端，以 openapi 的方式生成类型

"""

from email.policy import default
from typing import Literal
from pydantic import BaseModel


class DashNavItem(BaseModel):
    """
    菜单导航
    目前 以列表的方式表示，
    以后升级为树
    """

    title: str | None = None
    label: str | None = None
    icon: str | None = None
    url: str | None = None
    variant: Literal["ghost", "default"] | None = "default"
    tooltip: str | None = None


class DashConfig(BaseModel):
    """管理面板的后台配置"""

    logo: str | None = None
    pathPrefix: str | None = "/"
    navMenus: list[DashNavItem] | None = []
    loginUrl: str = "/auth/login"
    theme: str | None = "light"
    layout: str | None = "default"


class CmdkItem(BaseModel):
    """
    命令行菜单
    """

    label: str | None = None
    icon: str | None = None
    url: str | None = (
        None  # 用url的方式表示命令名称及参数, 例如： mtmai://local.mtmai.cmd?file_path=/data/data.json
    )



from sqlmodel import Field, SQLModel


class PageMetaAuthor(SQLModel):
    name: str = Field(default="")
    url: str = Field(default="")


class PageMetaBase(SQLModel):
    title: str | None = Field(default="")
    # 一般是URL的值
    metadataBase: str = Field(default="")
    description: str = Field(default="")
    keywords: list[str] = Field(default=[])
    authors: list[PageMetaAuthor] = Field(default=[])
    creator: str = Field(default="")
    # publisher: str = Field(default="")
    # copyright: str = Field(default="")
    # language: str = Field(default="")
    # image: str = Field(default="")
    # type: str = Field(default="")
    # url: str = Field(default="")
    manifest: str = Field(default="")  # eg. :`${siteConfig.url}/site.webmanifest`,


class PageMetaResponse(PageMetaBase):
    """对应前端页面的 page meta 信息"""

    pass



from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import Runnable, RunnableConfig

from mtmai.models.graph_config import InterviewState


writer_prompt = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            "You are an expert Wikipedia author. Write the complete wiki article on {topic} using the following section drafts:\n\n"
            "{draft}\n\nStrictly follow Wikipedia format guidelines.",
        ),
        (
            "user",
            'Write the complete Wiki article using markdown format. Organize citations using footnotes like "[1]",'
            " avoiding duplicates in the footer. Include URLs in the footer.",
        ),
    ]
)


class WriteArticleNode:
    def __init__(self, runnable: Runnable):
        self.runnable = runnable

    def node_name(self):
        return "write_article"

    async def __call__(
        self,
        state: InterviewState,
        config: RunnableConfig,
    ):
        topic = state["topic"]
        sections = state["sections"]
        draft = "\n\n".join([section.as_str for section in sections])

        writer = writer_prompt | self.runnable | StrOutputParser()
        article = await writer.ainvoke({"topic": topic, "draft": draft})
        return {
            **state,
            "article": article,
        }



from pathlib import Path

from mtmai.core.config import settings
from mtmai.core.logging import get_logger
from mtmlib.mtutils import bash

logger = get_logger()
easy_spider_repo = "https://github.com/NaiboWang/EasySpider"

easy_targt_dir = str(Path(settings.storage_dir, "easyspider"))


def install_easy_spider():
    if not Path(easy_targt_dir).exists():
        bash(f"git clone {easy_spider_repo} {easy_targt_dir}")
    else:
        bash(f"cd {easy_targt_dir} && git pull")

    bash(f"cd {easy_targt_dir} && bun i")


def run_easy_spider_server():
    logger.info(f"启动 easy spider in {easy_targt_dir}")
    if not Path(easy_targt_dir).exists():
        logger.warning("⚠️ Easy Spider not installed, installing now...")
        install_easy_spider()
        return

    bash(f"cd {easy_targt_dir} && bun run server.js")


def run_easy_spider_ui():
    logger.info(f"启动 easy spider in {easy_targt_dir}")
    if not Path(easy_targt_dir).exists():
        logger.warning("⚠️ Easy Spider not installed, installing now...")
        install_easy_spider()
        return

    bash(f"cd {easy_targt_dir} && bun run start_direct")




from datetime import time
import logging
from pathlib import Path

from mtmlib.mtutils import bash


logger = logging.getLogger()

# 官网： https://coder.com/docs/install
def start_code_server():
    config_file = Path.home().joinpath(".config/code-server/config.yaml")
    if not config_file.exists():
        logger.warning("code-server 配置文件不存在, 跳过启动: %s", config_file)
        return
    # 配置要点：
    # 1: 明确指定 SHELL 路径，否则在一些受限环境，可能没有默认的shell 变量，导致：“The terminal process "/usr/sbin/nologin" terminated with exit code: 1.”
    bash(
        "PORT=8622 PASSWORD=feihuo321 SHELL=/bin/bash code-server --bind-addr=0.0.0.0 &"
    )
    time.sleep(2)
    config_content = config_file.read_text()
    logger.info("codeserver 配置: %s", config_content)



# coding: utf-8

"""
    Mtmai API

    The Mtmai API

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictInt, StrictStr, field_validator
from typing import Any, ClassVar, Dict, List, Optional
from typing_extensions import Annotated
from mtmai.gomtmclients.rest.models.api_resource_meta import APIResourceMeta
from mtmai.gomtmclients.rest.models.recent_step_runs import RecentStepRuns
from mtmai.gomtmclients.rest.models.semaphore_slots import SemaphoreSlots
from mtmai.gomtmclients.rest.models.worker_label import WorkerLabel
from mtmai.gomtmclients.rest.models.worker_runtime_info import WorkerRuntimeInfo
from typing import Optional, Set
from typing_extensions import Self

class Worker(BaseModel):
    """
    Worker
    """ # noqa: E501
    metadata: APIResourceMeta
    name: StrictStr = Field(description="The name of the worker.")
    type: StrictStr
    last_heartbeat_at: Optional[datetime] = Field(default=None, description="The time this worker last sent a heartbeat.", alias="lastHeartbeatAt")
    last_listener_established: Optional[datetime] = Field(default=None, description="The time this worker last sent a heartbeat.", alias="lastListenerEstablished")
    actions: Optional[List[StrictStr]] = Field(default=None, description="The actions this worker can perform.")
    slots: Optional[List[SemaphoreSlots]] = Field(default=None, description="The semaphore slot state for the worker.")
    recent_step_runs: Optional[List[RecentStepRuns]] = Field(default=None, description="The recent step runs for the worker.", alias="recentStepRuns")
    status: Optional[StrictStr] = Field(default=None, description="The status of the worker.")
    max_runs: Optional[StrictInt] = Field(default=None, description="The maximum number of runs this worker can execute concurrently.", alias="maxRuns")
    available_runs: Optional[StrictInt] = Field(default=None, description="The number of runs this worker can execute concurrently.", alias="availableRuns")
    dispatcher_id: Optional[Annotated[str, Field(min_length=36, strict=True, max_length=36)]] = Field(default=None, description="the id of the assigned dispatcher, in UUID format", alias="dispatcherId")
    labels: Optional[List[WorkerLabel]] = Field(default=None, description="The current label state of the worker.")
    webhook_url: Optional[StrictStr] = Field(default=None, description="The webhook URL for the worker.", alias="webhookUrl")
    webhook_id: Optional[StrictStr] = Field(default=None, description="The webhook ID for the worker.", alias="webhookId")
    runtime_info: Optional[WorkerRuntimeInfo] = Field(default=None, alias="runtimeInfo")
    __properties: ClassVar[List[str]] = ["metadata", "name", "type", "lastHeartbeatAt", "lastListenerEstablished", "actions", "slots", "recentStepRuns", "status", "maxRuns", "availableRuns", "dispatcherId", "labels", "webhookUrl", "webhookId", "runtimeInfo"]

    @field_validator('type')
    def type_validate_enum(cls, value):
        """Validates the enum"""
        if value not in set(['SELFHOSTED', 'MANAGED', 'WEBHOOK']):
            raise ValueError("must be one of enum values ('SELFHOSTED', 'MANAGED', 'WEBHOOK')")
        return value

    @field_validator('status')
    def status_validate_enum(cls, value):
        """Validates the enum"""
        if value is None:
            return value

        if value not in set(['ACTIVE', 'INACTIVE', 'PAUSED']):
            raise ValueError("must be one of enum values ('ACTIVE', 'INACTIVE', 'PAUSED')")
        return value

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of Worker from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of metadata
        if self.metadata:
            _dict['metadata'] = self.metadata.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in slots (list)
        _items = []
        if self.slots:
            for _item_slots in self.slots:
                if _item_slots:
                    _items.append(_item_slots.to_dict())
            _dict['slots'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in recent_step_runs (list)
        _items = []
        if self.recent_step_runs:
            for _item_recent_step_runs in self.recent_step_runs:
                if _item_recent_step_runs:
                    _items.append(_item_recent_step_runs.to_dict())
            _dict['recentStepRuns'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in labels (list)
        _items = []
        if self.labels:
            for _item_labels in self.labels:
                if _item_labels:
                    _items.append(_item_labels.to_dict())
            _dict['labels'] = _items
        # override the default output from pydantic by calling `to_dict()` of runtime_info
        if self.runtime_info:
            _dict['runtimeInfo'] = self.runtime_info.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of Worker from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "metadata": APIResourceMeta.from_dict(obj["metadata"]) if obj.get("metadata") is not None else None,
            "name": obj.get("name"),
            "type": obj.get("type"),
            "lastHeartbeatAt": obj.get("lastHeartbeatAt"),
            "lastListenerEstablished": obj.get("lastListenerEstablished"),
            "actions": obj.get("actions"),
            "slots": [SemaphoreSlots.from_dict(_item) for _item in obj["slots"]] if obj.get("slots") is not None else None,
            "recentStepRuns": [RecentStepRuns.from_dict(_item) for _item in obj["recentStepRuns"]] if obj.get("recentStepRuns") is not None else None,
            "status": obj.get("status"),
            "maxRuns": obj.get("maxRuns"),
            "availableRuns": obj.get("availableRuns"),
            "dispatcherId": obj.get("dispatcherId"),
            "labels": [WorkerLabel.from_dict(_item) for _item in obj["labels"]] if obj.get("labels") is not None else None,
            "webhookUrl": obj.get("webhookUrl"),
            "webhookId": obj.get("webhookId"),
            "runtimeInfo": WorkerRuntimeInfo.from_dict(obj["runtimeInfo"]) if obj.get("runtimeInfo") is not None else None
        })
        return _obj





from typing import Any

from pydantic import BaseModel
from typing_extensions import Literal
from typing import Any, Dict, List, Optional

from pydantic.dataclasses import Field, dataclass



InputWidgetType = Literal[
    "string",
    "textarea",
    "number",
    "boolean",
    "tags",
    "array",
    "object",
    "select",
    "slider",
    "switch",
]


class InputWidgetBase(BaseModel):
    """
    输入组件的基类, 对应全段标准 html input 组件
    """

    class Config:
        arbitrary_types_allowed = True

    id: str | None = None
    name: str | None = None
    placeholder: str | None = None
    label: str | None = None
    tooltip: str | None = None
    description: str | None = None
    type: InputWidgetType | None = "string"
    placeholder: str | None = None
    value: str | None = None
    items: list[Dict[str, str]] = Field(default=[])


    # 附加选项，主要因为个别组件仅使用标准 html input 属性时不足以表达必要的参数
    options: dict[str, Any] | None = None

class ThreadForm(BaseModel):
    """发送Html 表单 向用户询问要执行任务的相关参数"""

    open: bool = True
    title: str = ""
    description: str = ""
    inputs: List[InputWidgetBase]
    disable_submit_button: bool = False
    display: Literal["modal", "drawer", "inline"] = "modal"
    variant: Literal["default", "cmdk","single_select"] = "default"


class TextInput(InputWidgetBase):
    """Useful to create a text input."""

    pass


class TextArea(InputWidgetBase):
    """Useful to create a text input."""

    type: InputWidgetType | None = "textarea"

class SelectInput(InputWidgetBase):
    """Useful to create a select input."""

    type: InputWidgetType = "select"
    initial: Optional[str] = None
    initial_index: int|None = None
    initial_value: str|None = None
    values: list[str] = Field(default=[])
    items: list[Dict[str, str]] = Field(default=[])
